import click 
from pypipet.core.logging import LEVELS
from pypipet.core.project_context import PipetContext

@click.group(invoke_without_command=True, no_args_is_help=True)
@click.option("--log-level", type=click.Choice(LEVELS.keys()), 
                             default='debug')
@click.option("--log-path", default='./')
@click.option("-v", "--verbose", count=True)
@click.option("-conf", "--config", default='setting.yaml')
@click.pass_context
def cli(ctx, log_level, log_path, verbose, config):
    """
    Get help at https://
    """
    #init context
    project = PipetContext()

    ctx.ensure_object(dict)
    ctx.obj["verbosity"] = verbose

    ctx.obj["config"] = config

    if ctx.obj.get("project") is None:
        ctx.obj["project"] = project
        if config == 'setting.yaml':
            ctx.obj["config"] = ctx.obj['project'].root + config

    if log_level:
        project.set_log_level(log_level, log_path)
        
    click.echo(project.log_level)
    


#!/usr/bin/python
#
# Copyright 2016 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Authentication-related info for the Content API for Shopping samples."""

from __future__ import print_function
import os
import sys

from google_auth_oauthlib import flow
from pypipet.plugins.gg_merchant.shopping.content import _constants
from pypipet.plugins.gg_merchant.shopping.content import token_storage
import google.auth
from google.oauth2 import service_account


def authorize(config):
  """Authorization for the Content API Samples.

  This function uses common idioms found across all the included
  samples, i.e., that service account credentials are located in a
  file called 'client-service.json' and that OAuth2 client credentials
  are located in a file called 'client-oauth2.json', both in the same
  directory as the sample configuration file.  Only one of these files
  needs to exist for authentication, and the service account will be
  chosen first if both exist.

  Args:
      config: dictionary, Python representation of config JSON.

  Returns:
      An google.auth.credentials.Credentials object suitable for
      accessing the Content API.
  """
  try:
    credentials, _ = google.auth.default(scopes=[_constants.CONTENT_API_SCOPE])
    print('Using application default credentials.')
    return credentials
  except google.auth.exceptions.DefaultCredentialsError:
    pass  # Can safely ignore this error, since it just means none were found.
  if 'path' not in config:
    print('Must use Application Default Credentials with no configuration.')
    sys.exit(1)
  service_account_path = os.path.join(config['path'],
                                      _constants.SERVICE_ACCOUNT_FILE)
  client_secrets_path = os.path.join(config['path'],
                                     _constants.CLIENT_SECRETS_FILE)
  if os.path.isfile(service_account_path):
    print('Using service account credentials from %s.' % service_account_path)
    return service_account.Credentials.from_service_account_file(
        service_account_path,
        scopes=[_constants.CONTENT_API_SCOPE])
  elif os.path.isfile(client_secrets_path):
    print('Using OAuth2 client secrets from %s.' % client_secrets_path)
    storage = token_storage.Storage(config)
    credentials = storage.get()
    if credentials and credentials.valid:
      return credentials
    client_config = token_storage.retrieve_client_config(config)
    auth_flow = flow.InstalledAppFlow.from_client_config(
        client_config, scopes=[_constants.CONTENT_API_SCOPE])
    credentials = auth_flow.run_local_server(authorization_prompt_message='')
    storage.put(credentials)
    return credentials
  print('No OAuth2 authentication files found. Checked:', file=sys.stderr)
  print('- Google Application Default Credentials', file=sys.stderr)
  print('- %s' % service_account_path, file=sys.stderr)
  print('- %s' % client_secrets_path, file=sys.stderr)
  print('Please read the accompanying documentation.', file=sys.stderr)
  sys.exit(1)
  return None



import os
from .db import project_engine, import_data
from .config_service import *
from sqlalchemy.orm import sessionmaker
from sqlalchemy.ext.automap import automap_base
from pypipet.plugins.woocommerce.api import API as WC_API
from pypipet.core.logging import setup_logging
from pypipet.core.fileIO.file_loader import read_yml_file
from pypipet.core.fileIO.file_saver import write_yaml
from .config import *
import logging
logger = logging.getLogger('__default__')

class PipetContext():
    def __init__(self, *args, **kwargs):
        self.engine = kwargs.get('engine', None)
        self.table_objects = kwargs.get('engine', None)
        self.root = kwargs.get('root_path', os.getcwd())
        if not self.root.endswith('/'):
            self.root += '/'
        self.config = None
        self.db_config = None
        self._shop_connectors = {}

    def start_project(self, project_name, **kwargs):
        if os.path.isfile(self.root + 'setting.yaml'):
            return False 
        project_setting = read_yml_file(PROJECT_SETTING)
        project_setting['project_setting']['name'] = project_name
        project_setting['project_setting']['home_dir'] = self.root
        # project_setting['project_setting']['attr_list'] = ['brand', 'upc', 'size', 'color']

        write_yaml(project_setting,
                   self.root + 'setting.yaml')

        os.makedirs(self.root + 'bundle', exist_ok=True)
        db_setting = read_yml_file(DB_SETTING)
        write_yaml(db_setting,
                   self.root + 'bundle/db_setting.yaml')

        wc_mapping = read_yml_file(WC_MAPPING)
        write_yaml(wc_mapping,
                   self.root + 'bundle/wc_field_mapping.yaml')
        file_template = read_yml_file(FILE_TEMPLATE)
        write_yaml(file_template,
                   self.root + 'bundle/file_template.yaml')
        return True

    def initialize_project(self, config_file=None):
        if self.config is None:
            self.config = read_yml_file(config_file)['project_setting']
       
        if self.db_config is None:
            self.db_config = read_yml_file(
                     self.config['home_dir'] \
                    + self.config['databse_setting'])
            is_valid, message = validate(self.db_config)

            if not is_valid:
                logger.info(message)
                return 
        
        if self.engine is None:
            self.engine = project_engine(self.db_config)

        if self.engine:
            self.set_table_objects(self.engine)
            self._session_maker = sessionmaker(bind=self.engine)

    def get_session_maker(self):
        if self.engine:
            if self._session_maker:
                return self._session_maker
            self._session_maker = sessionmaker(bind=self.engine)
            return self._session_maker
        return None
    
    def set_log_level(self, log_level, path='./'):
        self.log_level = log_level
        setup_logging(log_level, log_path=path)


    def set_table_objects(self, engine):
        base = automap_base()
        base.prepare(engine, reflect=True)
        self.table_objects = base.classes 

    def get_table_objects(self):
        return self.table_objects

    def get_table_columns(self, table_name):
        if self.table_objects.get(table_name):
            return self.table_objects.get(table_name).__table__.columns.keys()

    
    def set_shop_connector(self, name, connector):
        if connector.shop_type == 'wc':
            connector.set_field_mapping(self.get_wc_mapping())
        self._shop_connectors[name] = connector

    def get_shop_connector(self, name):
        return self._shop_connectors.get(name, None)

    def get_wp_api(self, shop_name: str):
        shop_config = self.config.get('shops').get(shop_name)
        if shop_config:
            return WC_API(
                    url=shop_config['url'],
                    consumer_key=shop_config['consumer_key'],
                    consumer_secret=shop_config['consumer_secret'],
                    version=shop_config['version']
                  )
    def set_wc_mapping(self):
        self._wc_mapping = get_wc_mapping(
                 filename=self.root + self.config['wc_field_mapping'])

    def get_wc_mapping(self):
        try:
            return self._wc_mapping
        except AttributeError as e:
            self.set_wc_mapping()
        return self._wc_mapping.get('mapping')

    def import_static_data(self):
        session = self._session_maker()
        settings = self.db_config
                
        if settings and settings.get('manual_data_input'):
            import_data(self.get_table_objects(), 
                       session, 
                       settings.get('manual_data_input'))
        session.close()
    

    def load_file_template(self):
        return read_yml_file(self.config['file_template'])
    
    


from pypipet.core.sql.query_interface import search_exist, update_data
from pypipet.core.sql.query_interface import add_json_to_db
import logging
from .utility import _object2dict
# from datetime import datetime, timedelta

logger = logging.getLogger('__default__')
logger.setLevel(logging.DEBUG)


def get_static_data(table_obj, session, params:dict={}):
    res = search_exist(table_obj, session, params)
    if res:
        return [_object2dict(r) for r in res]


def add_static_data(table_obj, session, data):
    return add_json_to_db(table_obj, session, data)

def edit_static_data(table_objs, session, data, target):
    if data.get('id') is not None:
        update_data(table_objs.get(target),
                    session,
                    data,
                    {'id': data['id']})
    if target == 'supplier' and data.get('name') is not None:
        update_data(table_objs.get(target),
                    session,
                    data,
                    {'name': data['name']}) 
    elif target == 'front_shop' and data.get('name') is not None:
        update_data(table_objs.get(target),
                    session,
                    data,
                    {'name': data['name']})
    elif target == 'tax' and data.get('name') is not None:
        update_data(table_objs.get(target),
                    session,
                    data,
                    {'name': data['name']}) 
    elif target == 'category' and data.get('category') is not None:
        update_data(table_objs.get(target),
                    session,
                    data,
                    {'category': data['category']}) 
    else:
        logger.debug('invalid input')

        


import os

CORE_DIR = os.path.abspath(os.path.dirname(__file__))
CORE_SETTINGS = f"{CORE_DIR}/default_setting"
PROJECT_SETTING =  f"{CORE_SETTINGS}/setting.yml"
DB_SETTING = f"{CORE_SETTINGS}/db_setting.yaml"
WC_MAPPING = f"{CORE_SETTINGS}/wc_field_mapping.yaml"
FILE_TEMPLATE = f"{CORE_SETTINGS}/file_template.yaml"




from pypipet.core.sql.model2query import *
from pypipet.core.sql.query import * 


def add_supplier_invoice(ctx, session, invoice: dict):
    if invoice.get('sku') is None:
        inv_data =  query_select(
            ctx.get_table_objects().get('inventory'), 
            session, 
            filters={'supplier_id': invoice['supplier_id'],
                    'supplier_item_id': invoice['supplier_item_id']})
        if len(inv_data) > 0:
            invoice['sku'] = inv_data[0].sku
        

    unique_keys = ctx.get_model_keys('supplier_invoice')
    add_if_not_exist(invoice, 
                    ctx.get_table_objects().get('supplier_invoice'), 
                        unique_keys,
                        session) 
    return True



#!/usr/bin/python
#
# Copyright 2017 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Run all example workflows (except for the Orders service)."""

from __future__ import absolute_import
from __future__ import print_function

import sys

from shopping.content import accounts
from shopping.content import accountstatuses
from shopping.content import accounttax
from shopping.content import common
from shopping.content import datafeeds
from shopping.content import products
from shopping.content import productstatuses
from shopping.content import shippingsettings


def main(argv):
  # Authenticate and construct service.
  service, config, _ = common.init(argv, __doc__)

  print('--------------------------------')
  accounts.workflow(service, config)
  print('--------------------------------')
  accountstatuses.workflow(service, config)
  print('--------------------------------')
  accounttax.workflow(service, config)
  print('--------------------------------')
  datafeeds.workflow(service, config)
  print('--------------------------------')
  products.workflow(service, config)
  print('--------------------------------')
  productstatuses.workflow(service, config)
  print('--------------------------------')
  shippingsettings.workflow(service, config)
  print('--------------------------------')


if __name__ == '__main__':
  main(sys.argv)



from sqlalchemy.sql import func
from sqlalchemy.exc import  SQLAlchemyError, IntegrityError
from sqlalchemy import select, update, desc
from sqlalchemy import and_, or_
from sqlalchemy.sql import text
from sqlalchemy.dialects.postgresql import insert as psql_insert
from pypipet.core.model.utility import get_unique_attrs
import json
from datetime import datetime, timedelta

import logging
logger = logging.getLogger('__default__')

#to do: pytest on this file

def object2dict(obj):
    d = {}
    for k in obj.__table__.columns.keys():
        d[k] = obj.__dict__.get(k, None)
    return d

def exception_handler(func):
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except  SQLAlchemyError as sqlError:
            kwargs.get('logger', logger).debug(sqlError)
    return wrapper

@exception_handler
def query_server_timestamp(session):
    statement = select(func.current_timestamp())
    results = session.execute(statement).first()
    return results

@exception_handler
def query_table_count(table_obj, session):
    count = session.query(table_obj).count()
    return count

def db_insert(obj, table_obj, session):
    """table_obj: sqlalchemy.ext.automap
       session: sqlalchemy session from session maker"""
    table_obj.created_at = func.now()
    table_obj.updated_at = func.now()

    try:
        rd = table_obj(**obj.get_all_attrs())
        session.add(rd)
        session.commit()
        session.refresh(rd)
        return rd
    except TypeError as e:
        logger.debug(e)
        session.rollback()
        logger.debug(json.dumps(obj.get_all_attrs()))
    except IntegrityError as e:
        # constraint on unique value
        logger.debug(f"integraty error {e}")
        session.rollback()
    except  SQLAlchemyError as sqlError:
        logger.debug(sqlError)
        session.rollback()
        logger.debug(json.dumps(obj.get_all_attrs()))

@exception_handler
def db_insert_raw(table_obj, session, data: dict):
    """table_obj: sqlalchemy.ext.automap
       session: sqlalchemy session from session maker"""
    data['created_at'] = func.now()
    data['updated_at']= func.now()
    try:
        rd = table_obj(**data)
        session.add(rd)
        session.commit()
        session.refresh(rd)
        return rd
    except TypeError as e:
        logger.debug(e)
        session.rollback()
        logger.debug(data)
    except IntegrityError as e:
        # constraint on unique value
        logger.debug(f"integraty error {e}")
        session.rollback()
    except  SQLAlchemyError as sqlError:
        logger.debug(sqlError)
        session.rollback()
        logger.debug(data)

@exception_handler
def db_insert_bulk(objs, table_obj, session):
    insert_list = []
    for obj in objs:
        obj.set_attr('created_at', str(func.now()))
        obj.set_attr('updated_at', str(func.now()))
        insert_list.append(obj.get_all_attrs())

    # logger.debug(f"bulk inserting {len(insert_list)}")
    session.bulk_insert_mappings(table_obj, insert_list)
    session.commit()

@exception_handler
def db_insert_raw_bulk(dict_list, table_obj, session):
    for i, _ in enumerate(dict_list):
        dict_list[i]['created_at'] = str(func.now())
        dict_list[i]['updated_at'] = str(func.now())

    # logger.debug(f"bulk inserting {len(dict_list)}")
    session.bulk_insert_mappings(table_obj, dict_list)
    session.commit()

@exception_handler
def db_select(table_obj, session, filters: dict = {}):
    statement = select(table_obj).filter_by(**filters)
    results = session.execute(statement).scalars().all()
    return results


@exception_handler
def table_latest_record(table_obj, session, key:str, filters: dict={}):
    statement = select(table_obj)\
        .filter_by(**filters)\
        .order_by(desc(table_obj.__table__.c[key]))
    latest = session.execute(statement).scalars().first()
    return latest


@exception_handler
def db_update(table_obj, session, update_data, filters: dict = {}):
    update_data['updated_at'] = str(func.now())
    statement = update(table_obj).\
                    filter_by(**filters).\
                    values(**update_data)
    # print(statement)
    session.execute(statement)
    session.commit()
    return update_data



def db_update_bulk(table_obj, session, data:list):
    dt = str(func.now())
    for d in data:
        d['updated_at'] = dt
    try:
        res = session.bulk_update_mappings(table_obj, data)
        session.commit()
        return res
    except IntegrityError as e:
        # constraint on unique value
        logger.debug(f"bulk update integraty error {e}")
        session.rollback()
    except  SQLAlchemyError as sqlError:
        logger.debug(sqlError)
        session.rollback()
        logger.debug(table_obj)

@exception_handler
def db_raw_query(query, session, params: dict={}):
    sql = text(query)
    # logger.debug(sql)
    results = session.execute(sql, params)
    data = []
    for r in results.fetchall():
        data.append(dict(r))
    return data


@exception_handler
def db_sum_query(table_obj, session, target, group_by:list, params:dict={}):
    statement = select(func.sum(target), 
                            *group_by)\
                .filter_by(**params)\
                .group_by(*group_by)
    res = session.execute(statement).all()
    return res

def get_products(table_objs, session, params:dict={}, details=0):
    product = table_objs.get('product')
     
    statement = select(table_objs.get('variation')).filter_by(**params)
    variation = statement.subquery()

    if details == 0:
        statement = select(product.product_name, 
                                variation.c.in_stock, variation.c.upc)\
                    .join_from(product, variation, 
                                variation.c.product_id==product.id)
    elif details == 1:
        statement = select(product.product_name, product.short_description, 
                     product.brand, product.category_id, variation)\
                    .join_from(product, variation, 
                                variation.c.product_id==product.id)
    res = session.execute(statement).all()

    return res

# @exception_handler
# def get_inventory(table_objs, session, start_sku, 
#                      batch_size=10, params: dict = {}, latest_hours=24):
#     inventory = table_objs.get('inventory')
#     variation = table_objs.get('variation')
#     dt = query_server_timestamp(session)
#     updated_after = dt[0]-timedelta(hours=latest_hours)
    
#     statement = None
#     if start_sku:
#         statement = select(variation.sku.label("sku"))\
#                     .filter(variation.sku > start_sku)\
#                     .order_by(variation.sku)\
#                     .limit(batch_size)
#     else:
#         statement = select(variation.sku.label("sku"))\
#                     .order_by(variation.sku)\
#                     .limit(batch_size)

#     if statement is None or session.execute(statement).scalars().first() is None:
#         return []
    
#     filtered_sku = statement.subquery()
    
#     statement = select(func.sum(inventory.qty).label("qty"), 
#                             inventory.sku.label("sku"))\
#                 .join_from(inventory, filtered_sku, 
#                             inventory.sku==filtered_sku.c.sku)\
#                 .filter_by(**params)\
#                 .filter(or_(inventory.updated_at > updated_after, 
#                                   inventory.created_at>updated_after))\
#                 .group_by(inventory.sku)\
#                 .order_by(inventory.sku)

#     # print(statement)
#     res = session.execute(statement).all()

#     return res

@exception_handler
def get_latest_inventory_update(table_obj, session, start_sku, 
                     batch_size=10, params: dict = {}, latest_hours=24):
    dt = query_server_timestamp(session)
    updated_after = dt[0]-timedelta(hours=latest_hours)
    statement = None 
    if start_sku is not None:
        statement = select(table_obj.sku)\
                    .distinct()\
                    .filter(table_obj.sku > start_sku)\
                    .filter_by(**params)\
                    .filter(or_(table_obj.updated_at > updated_after, 
                                    table_obj.created_at>updated_after))\
                    .order_by(table_obj.sku)\
                    .limit(batch_size)
    else:
        statement = select(table_obj.sku)\
                    .distinct()\
                    .filter_by(**params)\
                    .filter(or_(table_obj.updated_at > updated_after, 
                                    table_obj.created_at>updated_after))\
                    .order_by(table_obj.sku)\
                    .limit(batch_size)


    res = session.execute(statement).all()
    if len(res) == 0:
        return []

    filtered_sku = [r.sku for r in res]

    statement = select(func.sum(table_obj.qty).label('in_stock'), table_obj.sku)\
                    .filter(or_(table_obj.updated_at > updated_after, 
                                            table_obj.created_at>updated_after))\
                    .filter(table_obj.sku.in_(filtered_sku))\
                    .group_by(table_obj.sku)
    
    
    
    res = session.execute(statement).all()

    return res

@exception_handler
def get_variation_instock(table_objs, session, start_product_id, batch_size=20,
                               params: dict={}, latest_hours=24):
    variation = table_objs.get('variation')
    destination = table_objs.get('destination')
    dt = query_server_timestamp(session)
    updated_after = dt[0]-timedelta(hours=latest_hours)

    params.update({
        'available': True,
        'is_current_price': True
    })
    
    statement = None
    if start_product_id is None:
        statement = select(variation.in_stock.label("qty"), 
                            variation.sku.label("sku"), 
                            destination.destination_product_id.label("product_id"),
                            destination.destination_parent_id.label("parent_id"))\
                .join_from(variation, destination, 
                            variation.sku==destination.sku)\
                .filter_by(**params) \
                .filter(variation.updated_at > updated_after)\
                .filter(variation.in_stock >=0)\
                .filter(destination.destination_product_id != None)\
                .order_by(destination.destination_product_id)\
                .limit(batch_size)
    else:
        statement = select(variation.in_stock.label("qty"), 
                            variation.sku.label("sku"), 
                            destination.destination_product_id.label("product_id"),
                            destination.destination_parent_id.label("parent_id"))\
                .join_from(variation, destination, 
                            variation.sku==destination.sku)\
                .filter_by(**params) \
                .filter(variation.updated_at > updated_after)\
                .filter(destination.destination_product_id > start_product_id)\
                .filter(variation.in_stock >=0)\
                .filter(destination.destination_product_id != None)\
                .order_by(destination.destination_product_id)\
                .limit(batch_size)
    # print(statement)
    res = session.execute(statement).all()

    return res

@exception_handler
def get_fulfilled_order_in_processing(table_objs, session, status='processing', params:dict={}):
    statement = None
    if params.get('shop_order_id_list') is not None:
        sql = 'select * from fulfillment where shop_order_id=ANY(:shop_order_id_list)'
        return db_raw_query(sql, session, params={'shop_order_id_list': params['shop_order_id_list']}) 
    else:
        shop_order = table_objs.get('shop_order')
        fulfillment = table_objs.get('fulfillment')
        params.update({'status': status})
        statement = select(shop_order).filter_by(**params)
        orders = statement.subquery()

        statement = select(orders.c.id.label("shop_order_id"), 
                        orders.c.destination_order_id,
                        orders.c.front_shop_id, 
                        fulfillment.id, fulfillment.tracking_id, 
                        fulfillment.provider)\
                    .join_from(orders, fulfillment)
        res = session.execute(statement).all()
        
        return res

@exception_handler
def get_destination_parent(table_obj, session, front_shop_id, skus):
    sql = 'select destination_parent_id from destination where \
    front_shop_id=:front_shop_id and sku=ANY(:skus)'
    res = db_raw_query(sql, session, params={
        'front_shop_id': front_shop_id,
        'skus': skus
    })
    if res and len(res) >0 and res[0].get('destination_parent_id'):
        return res[0]['destination_parent_id']
# @exception_handler
# def query_select_raw(query, session, params: dict={}):
#     sql = text(query)
#     # logger.debug(sql)
#     results = session.execute(sql, params)
#     data = []
#     for r in results.fetchall():
#         data.append(dict(r))
#     return data

# @exception_handler
# def add_destination(data: dict, table_obj, unique_keys: list, session):
#     unique = get_unique_attrs(data, unique_keys)
#     unique['is_current_price'] = True
#     exist = query_select(table_obj, session, unique)
#     if exist and len(exist) > 0:
#         return exist[0]
#     # label old records is_current_price False 
#     if unique.get('price'):
#         del unique['price']

#     query_update(table_obj, session, 
#                 {'is_current_price': False}, 
#                 filters=unique)
#     data['is_current_price'] = True
#     return query_insert_from_data(data, table_obj, session)

# @exception_handler
# def query_variation(session, table_objs, params:dict={}):
#     variation = table_objs.get('variation')
#     destination = table_objs.get('destination')
#     res_vari = None
#     if params.get('sku'):
#         res_vari = query_select(variation, session, {'sku': params.get('sku')})

#     elif params.get('product_id'):
#         res_vari = query_select(variation, session, 
#                                {'product_id': params.get('product_id')})

#     if res_vari and len(res_vari) > 0:
#         res_vari = object2dict(res_vari[0])
#         res_dests = query_select(destination, session, {'sku': res_vari['sku']})
#         res_vari['destinations'] = [object2dict(d) for d in res_dests]
#         return res_vari
#     return None

# @exception_handler
# def query_order(session, table_objs, params:dict={}):
#     order = table_objs.get('shop_order')
#     order_item = table_objs.get('order_item')
#     customer = table_objs.get('customer')
#     res_order = None
#     if params.get('destination_order_id') and params.get('front_shop_id'):
#         res_order = query_select(order, session, 
#                       {'destination_order_id': params['destination_order_id'],
#                       'front_shop_id': params['front_shop_id']})
#     if res_order and len(res_order) > 0:
#         res_order = object2dict(res_order[0])

#         billing = query_select(customer, session, 
#                      {'id': res_order['billing_customer_id']})

#         if billing and len(billing) > 0:
#             res_order['billing_customer'] = object2dict(billing[0])

#         shipping = query_select(customer, session, 
#                      {'id': res_order['shipping_customer_id']})
#         if shipping and len(shipping) > 0:
#             res_order['shipping_customer'] = object2dict(shipping[0])

#         res_order_items = query_select(order_item, session, 
#                      {'shop_order_id': res_order['id']})
#         if res_order_items and len(res_order_items) > 0:
#             res_order['order_items'] = [object2dict(d) for d in res_order_items]

#         return res_order
#     return None


# @exception_handler
# def query_orders_fulfillment(session, table_objs, order_params: dict={}):
#     statement = None
#     if order_params.get('has_shipped'):
#         statement = select(table_objs.get('shop_order'),
#                        table_objs.get('shop_order').id,
#                        table_objs.get('shop_order').status,
#                        table_objs.get('shop_order').destination_order_id,
#                        table_objs.get('fulfillment').tracking_id,
#                        table_objs.get('fulfillment').provider)\
#         .join_from(table_objs.get('shop_order'),table_objs.get('fulfillment'))\
#         .filter(and_(table_objs.get('shop_order').status==order_params.get('status') ,
#                 table_objs.get('shop_order').front_shop_id==order_params.get('front_shop_id'))) 
#     else:
#         statement = select(table_objs.get('shop_order'))\
#         .outerjoin(table_objs.get('fulfillment'))\
#         .filter(and_(table_objs.get('shop_order').status==order_params.get('status') ,
#                 table_objs.get('shop_order').front_shop_id==order_params.get('front_shop_id') ,
#                 table_objs.get('fulfillment').id==None  ) ) 
               
#     res = session.execute(statement).all()
#     return res


# @exception_handler
# def gert_fulfillment_update(session, table_objs, updated_after: datetime, 
#                                                     params: dict = {}):
#     fulfillment = table_objs.get('fulfillment')
#     shop_order = table_objs.get('shop_order')
    
#     statement = select(shop_order.destination_order_id.label('destination_order_id'), 
#                          shop_order.payment_token.label('payment_token'), 
#                          fulfillment.tracking_id.label('tracking_id') )\
#                 .join_from(fulfillment, shop_order, 
#                             fulfillment.shop_order_id==shop_order.id)\
#                 .filter_by(**params) \
#                 .filter(fulfillment.created_at > updated_after)
    
#     res = session.execute(statement).all()
#     return res






# @exception_handler
# def get_fulfillment(session, table_objs, params: dict = {}):
#     statement = None
#     if params.get('shop_order_id'):
#         statement = select(table_objs.get('fulfillment'))\
#                    .filter(shop_order_id=params['shop_order_id'])
#     elif params.get('destination_order_id') and params.get('front_shop_id'):
#         statement = select(table_objs.get('shop_order'))\
#                    .filter_by(front_shop_id=params['front_shop_id'], 
#                               destination_order_id=params['destination_order_id'])
              
#         if session.execute(statement).scalars().first() is None:
#             return None, False          
#         statement = select(table_objs.get('fulfillment'))\
#                   .join(statement.subquery())


#     res = session.execute(statement).scalars().first()
    
#     if res:
#         return object2dict(res), True
    
#     return res, True


from copy import deepcopy

class BaseObject:
    def __init__(self, definition: dict = {}):
        self._definition = deepcopy(definition)

    def __getattr__(self, attr: str):
        if self._definition.get(attr):
            return self._definition.get(attr, None)
        else:
            return self.__dict__[attr]

    def __repr__(self):
        return f"<{self.__class__.__name__} >"
        

    def get_attr(self, name: str):
        return self._definition.get(name, None)

    def set_attr(self, name, value):
        if name:
            self._definition[name] = value

    def set_attrs(self, attr_names: list, attr_dict: dict):
        for attr_name in attr_names:
            self.set_attr(attr_name, attr_dict.get(attr_name, None))

    def re_map(self, remap:dict):
        if remap and len(remap) > 0:
            new_definition = {}
            for field, val in self._definition.items():
                new_definition[remap.get(field, field)] = val
            self._definition = new_definition

    def get_all_attrs(self):
        return self._definition

    def trim_attrs(self):
        for k, val in self._definition.items():
            if type(val) is str:
                self.set_attr(k, val.strip())

    def set_attrs_by_table_class(self, table_obj, attrs:dict):
        cols = list(table_obj.__table__.columns.keys())
        for col in cols:
            if attrs.get(col) is not None:
                self.set_attr(col, attrs[col])


    def validate_fields(self, table_obj):
        cols = list(table_obj.__table__.columns.keys())
        validated = {}
        for col in cols:
            if self._definition.get(col) is not None:
                validated[col] = self._definition[col]
        self._definition = validated

    def get_unique_attrs(self, unique_keys: list):
        unique_attrs = {}
        for k in unique_keys:
            unique_attrs[k] = self.get_attr(k)
        return unique_attrs


    



"""woocommerce version 5.5.2"""

import logging
logger = logging.getLogger('__default__')

from pypipet.core.shop_conn.wc import transform_imgs, transform_category, transform_attrs, transform_attrs_variations

_DEBUG_ = True

def exception_handler(func):
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except  TypeError as e:
            logger.debug(e)
    return wrapper


def parse_to_wp_product_variable(shop_api, product_info: dict, attr_list,
                        variation_attrs: list, update_only=False, parent_id=None):
    if product_info.get('variations') is None \
    or len(product_info['variations']) == 0:
        return  
    
    addtional_attrs = {}
    if product_info.get('brand'): addtional_attrs['brand'] = product_info['brand']
    attrs = transform_attrs_variations(shop_api, attr_list, 
                               product_info['variations'], variation_attrs, 
                               addtional_attrs=addtional_attrs)
    
    parent_product = {'name': product_info['product_name'],
                    'short_description': product_info['short_description'],
                    'type': 'variable',
                    'attributes': attrs,
                    'default_attributes': product_info['variations'][0]['attributes'],
                    'description': product_info['variations'][0]['description'],
                    # 'category': product_info['category']
                    }
    if parent_product.get('attributes') is not None \
    and len(parent_product['attributes']) == 0:
        del parent_product['attributes']
    if parent_product.get('default_attributes') is not None and \
    len(parent_product['default_attributes']) == 0:
        del parent_product['default_attributes']
    # if product_info.get('category'):
    #     parent_product['category'] = product_info['category']
    
    all_images = []
    for vari in product_info['variations']:
        if vari.get('images'):
            imgs = transform_imgs(vari['images'])
            all_images += imgs
            vari['image'] = imgs[0]
            # del vari['images']
        if vari.get('sub_title'):
            vari['description'] = f"sku {vari['sku']} {vari['sub_title']}"
            del vari['sub_title']

    
    if _DEBUG_: parent_product['images'] = all_images[:2]

    if parent_id is None and not parse_to_wp_product(shop_api, parent_product, 
                                          attr_list, product_type='variable', 
                                          update_only=False, is_variation=False):
        return 

    for i, vari in enumerate(product_info['variations']):
        parse_to_wp_product(shop_api, vari, attr_list, product_type='variable', 
                                          update_only=False, is_variation=True)
        if vari.get('images'): del vari['images']                                  
        product_info['variations'][i] = vari
    
    return {'parent': parent_product, 'variations': product_info['variations']}
        
    # return attrs, product_info['variations']
   
def parse_to_wp_product(shop_api, product_info: dict, attr_list, product_type='simple', 
                                   status='publish',  update_only=False, is_variation=False):
    if not _validate_product_info(product_info, 
                        is_parent=product_type=='variable' and not is_variation,
                        update_only=update_only): 
        return False
        
    if not is_variation and product_type != 'variable':
        product_info['attributes']  = transform_attrs(shop_api, 
                                                    attr_list, 
                                                    product_info)
    else:
        for attr in attr_list:
            if product_info.get(attr): del product_info[attr]

    if product_info.get('attributes') is not None \
    and len(product_info['attributes']) == 0:
        del product_info['attributes']

    if product_info.get('product_name') is not None:
        product_info['name'] = product_info['product_name']
        del product_info['product_name']
    
    if product_info.get('discount') is not None :
        if product_info['discount'] == 0:
            product_info['sale_price'] = ''
        else:
            product_info['sale_price'] = product_info['price'] \
                                            - float(product_info['discount'])
                                           
    if product_info.get('price') is not None:
        product_info['regular_price'] = str(round(float(product_info['price']),2))
        del product_info['price']
    
    if product_info.get('images') and not is_variation and product_type != 'variable':
        product_info['images'] =  transform_imgs(product_info['images'])
        if _DEBUG_: product_info['images'] = product_info['images'][:2]

    if product_info.get('category') is not None:
        product_info['categories'] = [transform_category(shop_api, 
                                            product_info['category'])]
        del product_info['category']

    if product_info.get('weight') is not None:
        product_info['weight'] = str(product_info['weight'])

    if product_info.get('length') is not None and product_info.get('width') \
        is not None and product_info.get('height') is not None:
            _transform_dimension(product_info)
    
    if not update_only:
        if not is_variation:
            product_info.update({
                'type': product_type,
                'status': status
            })
        else:
            if product_info.get('product_id'):
                del product_info['product_id']
        if product_info.get('id') is not None:
            del product_info['id']
    else:
        if product_info.get('product_id') is not None:
            product_info['id'] = int(product_info['product_id'])
            del product_info['product_id']
  
    
    if product_info.get('available') is not None\
    and product_info['available'] == False:
        product_info['in_stock'] = 0

    if product_info.get('in_stock') is not None:
        product_info.update(transform_inventory(product_info['in_stock']))
        del product_info['in_stock']

    if product_info.get('qty') is not None:
        product_info.update(transform_inventory(product_info['qty']))
        del product_info['qty']

    if product_info.get('updated_at') is not None:
        del product_info['updated_at']
    if product_info.get('created_at') is not None:
        del product_info['created_at']
    
    return True

# def object_to_wp_product(product_info: dict, product_type='simple', update_only=False):
#     if product_info.get('product_name') is not None:
#         product_info['name'] = product_info['product_name']
#         del product_info['product_name']

#     if product_info.get('price') is not None:
#         product_info['regular_price'] = str(round(float(product_info['price']),2))
#         del product_info['price']

#     if product_info.get('category') is not None:
#         product_info['categories'] = product_info['category']
#         del product_info['category']

#     if product_info.get('weight') is not None:
#         product_info['weight'] = str(product_info['weight'])
    
#     if not update_only:
#         product_info.update({
#             'type': product_type,
#             'status': 'publish'
#         })
  
#     if product_info.get('discount') is not None :
#         product_info['sale_price'] = product_info['price'] \
#                                             - float(product_info['discount'])
#     if product_info.get('available') is not None\
#     and product_info['available'] == False:
#         product_info['in_stock'] = 0

#     if product_info.get('in_stock') is not None:
#         product_info.update(transform_inventory(product_info['in_stock']))
#         del product_info['in_stock']
#     if not update_only and product_info.get('id') is not None:
#         del product_info['id']

#     if product_info.get('updated_at') is not None:
#         del product_info['updated_at']
#     if product_info.get('created_at') is not None:
#         del product_info['created_at']


def transform_inventory(qty: int):
    inv = {
         'manage_stock': True,
         'stock_quantity': qty,
    }
    if qty == 0:
        inv['stock_status'] = 'outofstock'
    else:
        inv['stock_status'] = 'instock'
    # print(inv)
    return inv

# def transform_batches(batch_data: list):
#     for i, d in enumerate(batch_data):
#         if d.get('product_id') is None and d.get('id'):
#             continue 

#         if d.get('product_id'):
#             batch_data[i]['id'] = int(d['product_id'])
#             del batch_data[i]['product_id']
#         elif d.get('id'):
#             batch_data[i]['id'] = int(d['id'])
        
#         if d.get('qty') is not None:
#             batch_data[i].update(transform_inventory(int(d['qty'])))
#             del batch_data[i]['qty']

def transform_customer(data):
    if data.get('address1') is not None:
        data['addres_1'] = data['address1']
    if data.get('address2') is not None:
        data['addres_2'] = data['address2']
    if data.get('updated_at') is not None:
        del data['updated_at']
    if data.get('created_at') is not None:
        del data['created_at']

def transform_order_item(order:dict, items:dict):
    if order.get('line_items') is None:
        return 

    item_list = []
    for item in order['line_items']:
        if items.get(item['product_id']) is None:
            continue
        
        if items[item['product_id']].get('qty'):
            item['quantity'] = items[item['product_id']]['qty']
        elif items[item['product_id']].get('order_qty'):
            item['quantity'] = items[item['product_id']]['order_qty']
        elif items[item['product_id']].get('ship_qty'):
            item['quantity'] = items[item['product_id']]['ship_qty']

        item['subtotal'] = str(round(items[item['product_id']]['price']*item['quantity'], 2)) 
        item['total'] = float(item['subtotal']) + float(item['total_tax'])
        item['total'] = str(round(item['total'], 2)) 
        print(item)
        item_list.append(item)

    return item_list
        

def _validate_product_info(product_info, is_parent=False, update_only=False):
    if not update_only and product_info.get('images') is None:
        logger.debug('missing images info')
        return False

    # if product_info.get('price') is None:
    #     logger.debug('missing price info')
    #     return False

    if not update_only and not is_parent and product_info.get('sku') is None:
        logger.debug('missing sku info')
        return False

    return True

def _transform_dimension(product_info):
    product_info['dimensions'] =  {
        'length': str(product_info['length']),
        'width': str(product_info['width']),
        'height': str(product_info['height'])
    }





import logging, time, re
logger = logging.getLogger('__default__')

_MAX_RETRY =3

"""
note: wc can not search precisely with name
"""

def request_handler(func):
    def wrapper(*args, **kwargs):
        try:
            res = func(*args, **kwargs)
            if res and res.status_code in [200, 201]:
                return res.json()
            if res:
                logger.debug(res.text)
            return None
        except  KeyError as e:
            logger.debug('requst error')
            logger.debug(e)
    return wrapper

def request_handler_retry(func):
    def wrapper(*args, **kwargs):
        for i in range(_MAX_RETRY):
            try:
                res = func(*args, **kwargs)
                if res.status_code in [200, 201]:
                    return res.json()
                logger.debug(res.text)
            except  KeyError as e:
                logger.debug('requst error')
                logger.debug(e)
        return None
    return wrapper

# def validate_by_key(key):
#     def decorator(func):
#         def wrapper(*args, **kwargs):
#             try:
#                 res = func(*args, **kwargs)
#                 if res and kwargs.get(key):
#                     for r in res:
#                         if r.get(key) == kwargs.get(key):
#                             return r 
#                 return None

#             except  KeyError as e:
#                 logger.debug('validate error')
#                 logger.debug(e)
#         return wrapper
#     return decorator


#
#------function of handling product-------------
#

@request_handler
def search_product_by_sku(wcapi, sku:str):
    return wcapi.get('products', params={'sku': sku})

def get_product_by_sku(wcapi, sku:str):
    res = search_product_by_sku(wcapi, sku)
    for p in res:
        if p['sku'] == sku:
            return p

# @request_handler
def add_product(wcapi, product_info):
    res =  wcapi.post('products', data=product_info)
    if res is None: return 
    if res.status_code in [200, 201]:
        return res.json()
    
    if res and res.json().get('code'):
        error_code = res.json()['code']
        if 'sku' in error_code and 'duplicate' in error_code:
            #update product
            exist_product = get_product_by_sku(wcapi, product_info['sku'])
            if exist_product is not None:
                res =  update_product(
                            wcapi,
                            str(exist_product['id']) ,
                            product_info, 
                            )
                if res and res.status_code in [200, 201]:
                    return res.json()

    logger.debug(res.text)

# @request_handler
def add_product_variations(wcapi, product_info, parent_id=None):
    if parent_id is None:
        res = wcapi.post('products', data=product_info['parent']) 
        if res is None: return 
        if res.status_code not in [200, 201]: 
            logger.debug(res.text)
            return 
        parent_id = res.json()['id']
    else:
        parent_exist = wcapi.get(f'products/{parent_id}')
        if parent_exist.status_code not in (201, 200):
            return 
        parent_update = _merge_parent_attrs(product_info['variations'], 
                                          parent_exist.json()['attributes'])
        wcapi.put(f'products/{parent_id}', data={'attributes': parent_update}) 

    if parent_id is not None:
        logger.debug(f'parent_id {parent_id}')
        for i, vari in enumerate(product_info['variations']):
            # note: this is not in woocomerce documents
            product_info['variations'][i]['product_id'] = parent_id
        
        batch_data = {'create': product_info['variations']}
        res = wcapi.post(f'products/{parent_id}/variations/batch', data=batch_data)
        
        if res is not None and res.status_code in [200, 201]: 
           return  _parse_variation_response(res.json()['create'], parent_id)
        else:
            logger.debug(res.status_code)
            logger.debug(res.json())

def _merge_parent_attrs(variations, prev_attrs):
    addtional_attrs = {}
    for vari in variations:
        for attr in vari['attributes']:
            if addtional_attrs.get(attr['id']) is None:
                addtional_attrs[attr['id']] = []
            addtional_attrs[attr['id']].append(attr['option'])

    for i, attr in enumerate(prev_attrs):
        if addtional_attrs.get(attr['id']) is not None:
            prev_attrs[i]['options']+= addtional_attrs[attr['id']]

    return prev_attrs

def _parse_variation_response(vari_list: list, parent_id):
    res = {}
    for i, vari in enumerate(vari_list):
        if vari.get('sku') is None: 
            logger.debug(f'wc bulk add failed {vari}')
            continue
        res[vari['sku']] = {
                'id': vari['id'],
                'parent_id': parent_id}
    return res

@request_handler
def get_product_by_id(wcapi, pid:int):
    return wcapi.get(f'products/{pid}')

@request_handler
def list_products(wcapi,  params: dict={}):
    """start from page 1
       by default woocommerce order orders by order_id DESC
    """
    params.update({
        'orderby': 'id',
        'order': 'asc'
    })
    return wcapi.get('products', params=params)


@request_handler
def update_product(wcapi, pid:str, data:dict):
    if data.get('regular_price') and type(data['regular_price']) is not str:
        data['regular_price'] = str(round(data['regular_price'], 2))
    if data.get('sale_price') and type(data['sale_price']) is not str:
        data['sale_price'] = str(round(data['sale_price'], 2))
    # print(data, pid)
    return wcapi.put(f'products/{pid}', data=data) 

@request_handler_retry
def update_product_batch(wcapi, action:str, data:list):
    batch_data = {action: data}
    return wcapi.put(f'products/batch', data=batch_data) 

@request_handler
def add_variation(wcapi, product_id, variation_info):
    wcapi.post(f'products/{product_id}/variations', data=variation_info)

@request_handler
def list_variations(wcapi, product_id, params: dict ={}):
    params.update({
        'orderby': 'id',
        'order': 'asc'
    })
    return wcapi.get(f'products/{product_id}/variations', params=params)

@request_handler
def get_variation(wcapi, product_id, variation_id):
    return wcapi.get(f'products/{product_id}/variations/{variation_id}')

@request_handler
def update_variation(wcapi, product_id, variation_id, data):
    if data.get('regular_price') and type(data['regular_price']) is not str:
        data['regular_price'] = str(round(data['regular_price'], 2))
    if data.get('sale_price') and type(data['sale_price']) is not str:
        data['sale_price'] = str(round(data['sale_price'], 2))
    
    variant = get_variation(wcapi, product_id, variation_id)
    if data.get('attributes') is None or len(data['attributes']) == 0:
        data['attributes'] = variant['attributes']
    else:
        data['attributes'] += variant['attributes']
    return wcapi.put(f'products/{product_id}/variations/{variation_id}', data)

# @request_handler
# def update_variations_batch(wcapi, product_id,  action:str, data:list):
#     batch_data = {action: data}
#     return wcapi.post(f'products/{product_id}/variations/batch', data=batch_data)

@request_handler
def list_categories(wcapi, page:int, per_page:int =10, params: dict={}):
    """start from page 1
       by default woocommerce order orders by order_id DESC
    """
    params.update({
        'page': page,
        'per_page': per_page,
        'orderby': 'id',
        'order': 'asc'
    })
    return wcapi.get('products/categories', params=params)



#
#------function of handling orders-------------
#



@request_handler
def list_orders(wcapi, page:int, per_page:int =10):
    return wcapi.get('orders', params={'page': page,
                                         'per_page': per_page})

@request_handler
def get_order_by_id(wcapi, oid):
    return wcapi.get(f'orders/{oid}')

def get_new_orders(wcapi, start_from_order:str=None, 
                     page_start:int = 1, per_page:int =10):
    """
    sync new orders from wp 
    start_from_order: latest inserted order id 
    page_start: fetch from page number
    """
    orders = []
    page = page_start
    while True:
        logger.debug(f'fetching order page {page}')
        batch = list_orders(wcapi, page, per_page=per_page)
        if batch is None: return None
        if len(batch) == 0: break
        for order in batch:
            if start_from_order is not None and \
            str(order['id']) == start_from_order.strip():
                return orders
            orders.insert(0,order)
        page += 1 

    return orders

@request_handler
def send_note(wcapi, order_id:str, data:dict):
    return wcapi.post(f'orders/{order_id}/notes', data=data)
    

def send_wc_message(wcapi, order_id: str, template: str):
    res = send_note(wcapi, order_id, data={
        'customer_note': True,
        'note': template
    })
    return res

@request_handler
def update_order(wcapi, order_id, data):
    return wcapi.put(f'orders/{order_id}', data)

# @request_handler
# def update_customer(wcapi, customer_id, data):
#     print(data, customer_id)
#     return wcapi.put(f'customers/{customer_id}', data)

#
#------function of transforming product attributes-------------
#

def transform_imgs(image_str):
    try:
        images = image_str.split(',')
        return [{'src': img.strip(), "position": i}
                for i, img in enumerate(images)]
    except:
        return []

def transform_categories(wcapi, full_path, sep='>'):
    cats = full_path.split('>')
    parent_id = None
    categories = []
    for c in cats:
        new_cat = transform_category(wcapi, c, parent_id=parent_id)
        assert new_cat is not None
        categories.append(new_cat)
        # parent = new_cat
    return categories

def transform_category(wcapi, category, parent_id=None):
    category = get_category(wcapi, category.strip(), parent_id=parent_id)
    if category is None:
        return None 
    return {'id': category['id']}


def get_category(wcapi, name, parent_id=None):
    category = search_category(wcapi, name)
    if category is not None:
        return category

    data  = {'name': name}
    if parent_id:
        data['parent'] = parent_id
    return add_category(wcapi, data)

@request_handler
def add_category(wcapi, data):
    return  wcapi.post('products/categories', data=data)

def search_category(wcapi, name):
    name_slug = re.sub('[^a-zA-Z0-9_]+', '-', name.replace('\'', ''))
    # print(name, name_slug)
    res = wcapi.get('products/categories', params={'search': name_slug})
    if res.status_code in [200, 201]:
        for cat in res.json():
            if cat['slug'] == name_slug.lower():
                return cat
    return None

@request_handler
def add_brand(wcapi, name):
    return wcapi.post('products/brands', data={'name': name})
    

def get_brand(wcapi, name):
    name_slug = re.sub('[^a-zA-Z0-9]+', '-', name.lower())
    res = wcapi.get('products/brands', params={'search': name_slug})
    if res.status_code in [200, 201]:
        for item in res.json():
            if item['slug'] == name_slug:
                return item
        return add_brand(wcapi, name)
    else:
        logger.debug(res.text)
        return None 

def transform_brand(wcapi, name):
    if name:
        brand = get_brand(wcapi, name)
        if brand is not None:
            return [brand['id']]

# def get_tags(wcapi, names):
#     tags = []
#     for name in names:
#         t = getTag(wcapi, name.strip().lower())
#         # assert t is not None
#         if t is None: continue
#         tags.append({'id': t})
#     return tags

@request_handler
def add_attr(wcapi, name):
    return wcapi.post("products/attributes", data={
        'name': name
    })


def get_attr(wcapi, name):
    res = wcapi.get('products/attributes', params={'search': name})
    if res.status_code in [200, 201]:
        for item in res.json():
            if item['name'].lower() == name.lower():
                return item
        return add_attr(wcapi, name)
    logger.debug(res.text)
    return None 

def transform_attrs(wcapi, attrs_list, product_info):
    attrs = []
    for attr_name in attrs_list:
        if product_info.get(attr_name) is None \
                   or str(product_info[attr_name]).strip() == "": 
            continue
        attr = get_attr(wcapi, attr_name)
        if attr is  None: continue

        attrs.append({
            'id' : attr['id'], 
            # 'position' : 0,
            'visible' : True,
            'options' : [
                str(product_info[attr_name])]
        })

       
    return attrs



def transform_attrs_variations(wcapi, attrs_list, variations, 
                                        variation_attrs, addtional_attrs={}):
    parent_attrs = {}
    for attr_name in attrs_list:
        attr = get_attr(wcapi, attr_name)
        if attr is  None: continue
            
        if parent_attrs.get(attr_name) is None:
            parent_attrs[attr_name] = {
                'id' : attr['id'], 
                'visible' : True,
                'options' : []
            }


    for vari in variations:
        vari['sub_title'] = ''
        attrs = []
        for attr_name in attrs_list:
            if vari.get(attr_name) is None \
                    or str(vari[attr_name]).strip() == "": 
                continue

            if parent_attrs.get(attr_name) is None:
                continue
            
            if attr_name in variation_attrs:
                attrs.append({
                    'id' : parent_attrs[attr_name]['id'], 
                    # 'position' : 0,
                    'option' : str(vari[attr_name])
                })
                vari['sub_title'] += vari[attr_name] + ' '
                parent_attrs[attr_name]['variation'] = True
         
            if len(parent_attrs[attr_name] ['options']) == 0 \
            or parent_attrs[attr_name].get('variation') == True \
            or attr_name.lower() in ('upc', 'mpn', 'din'):
                parent_attrs[attr_name]['options'].append(str(vari[attr_name]))
        vari['attributes'] = attrs
    
    parent_attrs.update(addtional_attrs)
    empty_attrs = []
    for k, val in parent_attrs.items():
        if val and( type(val) is str or val.get('id') is None):
            attr = get_attr(wcapi, k)
            if attr is  None: continue
                
            parent_attrs[k] = {
                    'id' : attr['id'], 
                    'visible' : True,
                    'options' : [str(val)]
                }   
        elif val and len(val['options']) == 0:
            empty_attrs.append(k)

    for attr in empty_attrs:
        del parent_attrs[attr]
       
    return list(parent_attrs.values())

# def transform_dimension(wcapi, product_info):
#     product_info['dimensions'] =  {
#         'length': str(product_info['length']),
#         'width': str(product_info['width']),
#         'height': str(product_info['height'])
#     }


"""Base Error classes."""
import functools
import io
import logging
# import subprocess
# from asyncio.streams import StreamReader
# from asyncio.subprocess import Process
# from enum import Enum
# from typing import Optional, Union

def aggregate(error_cls):
    class Aggregate(error_cls):
        """Aggregate multiple sub-exceptions."""

        def __init__(self, exceptions: []):
            self.exceptions = exceptions

        def __str__(self):
            return "\n".join(str(e) for e in self.exceptions)

    if error_cls != Exception:
        error_cls.Aggregate = Aggregate

    return error_cls


AggregateError = aggregate(Exception)

class DialectNotSupportedError(Exception):
    def exit_code(self):
        return ExitCode.FAIL



import pypipet.core.shop_conn.wc as wc
import pypipet.core.transform.wc_to_model as wc_to_model
import pypipet.core.transform.model_to_wc as model_to_wc

import logging
logger = logging.getLogger('__default__')

def _clean_up_data(data):
    keys = set(list(data.keys()))
    if 'created_at' in keys:
        del data['created_at']
    if 'updated_at' in keys:
        del data['updated_at']


def validate_product_info(product_info):
    if product_info.get('images') is None:
        logger.debug('missing images info')
        return False

    if product_info.get('price') is None:
        logger.debug('missing price info')

    if product_info.get('sku') is None:
        logger.debug('missing sku info')

    return True

class ShopConnector():
    def __init__(self, shop_name, shop_type, **kwargs):
        self.shop_name = shop_name
        self.shop_type = shop_type
        self.shop_api = kwargs.get('api', None)
        self.front_shop_id = kwargs.get('front_shop_id', None)
        self.field_mapping = kwargs.get('field_mapping', None)
        self.batch_size = kwargs.get('batch_size', 10)

    def set_shop_api(self, api):
        self.shop_api = api 

    def set_front_shop_id(self, shop_id):
        self.front_shop_id = shop_id

    def set_field_mapping(self, mapping):
        self.field_mapping = mapping

    def get_shop_info(self):
        return {
            'shop_name': self.shop_name,
            'shop_type': self.shop_type,
            'front_shop_id': self.front_shop_id
        }

    @staticmethod
    def add_product_to_shop(shop_type, shop_api, product_info, **kwargs):
        # _clean_up_data(product_info)
        if shop_type == 'wc':
            attr_list = kwargs.get('attr_list', ['brand', 'upc', 'size', 'color'])
            if kwargs.get('variation_attrs') is None or product_info.get('variations') is None:
                if model_to_wc.parse_to_wp_product(shop_api, product_info, attr_list, 
                                         status=kwargs.get('publish_type', 'publish')):
                    return wc.add_product(shop_api, product_info)
            else:
                product = model_to_wc.parse_to_wp_product_variable(shop_api, product_info, 
                                              attr_list,kwargs['variation_attrs'],
                                              parent_id=kwargs.get('parent_id'))
                
                if product:
                    return wc.add_product_variations(shop_api, product, parent_id=kwargs.get('parent_id'))

        else:
            logger.debug('missing shop info')

    @staticmethod
    def add_product_to_shop_bulk(shop_type, shop_api, products:list,  **kwargs):
        """for woocommerce, bulk add only for non-variable product"""
        if shop_type == 'wc':
            attr_list = kwargs.get('attr_list', ['brand', 'upc', 'size', 'color'])
            batch = []
            for product_info in products:
                if model_to_wc.parse_to_wp_product(shop_api, product_info, attr_list):
                    batch.append(product_info)
            #batch add 
            res = wc.update_product_batch(shop_api, 'create', batch)

        else:
            logger.debug('missing shop info')


    @staticmethod
    def get_destination_product_id(shop_type, shop_api, sku:str):
        """
        shop: dict with keys (front_shop_id, shop_api, shop_type)
        """
        if shop_type == 'wc':
            product = wc.get_product_by_sku(shop_api, sku)
            if product:
                return product['id']

        return None

    @staticmethod
    def update_shop_product(shop_type, shop_api, data:dict, product_id:str, **kwargs ):
        if shop_type == 'wc':
            attr_list = kwargs.get('attr_list', ['brand', 'upc', 'size', 'color'])
            if kwargs.get('formated') is None or kwargs['formated'] == False:
                if not model_to_wc.parse_to_wp_product(shop_api, data, attr_list, update_only=True):
                    logger.debug('wp parsing failed')
                    return 
            # print(data)
            if kwargs.get('parent_id') is not None:
                # print(kwargs['parent_id'], data)
                return wc.update_variation(shop_api, kwargs['parent_id'], product_id, data)
            return wc.update_product(shop_api, product_id, data)

    @staticmethod
    def update_shop_product_batch(shop_type, shop_api, data:list, **kwargs):
       if shop_type == 'wc':
            attr_list = kwargs.get('attr_list', ['brand', 'upc', 'size', 'color'])
            batch = []
            for i, d in enumerate(data):
                # print(d)
                if kwargs.get('formated') is None or kwargs['formated'] == False:
                    if model_to_wc.parse_to_wp_product(shop_api, d, attr_list, update_only=True):
                        if d.get('parent_id') is not None:
                            # print(d)
                            wc.update_variation(shop_api, d['parent_id'] , d['id'], d)
                        else:
                            batch.append(d)
                else:
                    batch.append(d)
            
            res = wc.update_product_batch(shop_api, 'update', batch)
            return res

    @staticmethod
    def sync_shop_orders(shop_type, shop_api, field_mapping, **kwargs):
        res = []
        if shop_type == 'wc':
            wp_orders = wc.get_new_orders(
                shop_api,
                start_from_order=kwargs.get('latest_order_id', '0'), 
                page_start=kwargs.get('page_start', 1), 
                per_page=kwargs.get('per_page', 20))
        
            for order in wp_orders:
                print('prcessing order', order['id'])
                # todo: get tax based on geo
                data = wc_to_model.wp_parse_order(order, field_mapping, 
                                shipping_tax_id=kwargs.get('shipping_tax_id',2), 
                                item_tax_id=kwargs.get('item_tax_id', 1)) 
                res.append(data)

        return res


    @staticmethod
    def update_order_status_at_shop(shop_type, shop_api, destination_order_id, 
                                   status,**kwargs):
        if shop_type == 'wc':
            res = wc.update_order(shop_api,
                                destination_order_id,
                                data={'status': status})
            return res

    @staticmethod
    def update_order_at_shop(shop_type, shop_api, destination_order_id, 
                             data, **kwargs):
        if shop_type == 'wc':
            res = wc.update_order(shop_api,
                                destination_order_id,
                                data=data)
            return res

    @staticmethod
    def update_order_item_at_shop(shop_type, shop_api, destination_order_id, 
                             items, **kwargs):
        if shop_type == 'wc':
            order = ShopConnector.get_order_at_shop(shop_type, shop_api,
                            destination_order_id=destination_order_id)
            item_list = model_to_wc.transform_order_item(order, items)
            res = wc.update_order(shop_api,
                                destination_order_id,
                                data={'line_items': item_list})
            return res

    @staticmethod
    def get_order_at_shop(shop_type, shop_api, **kwargs):
        # print(kwargs.get('destination_order_id'))
        if shop_type == 'wc':
            res = wc.get_order_by_id(shop_api,
                                kwargs.get('destination_order_id'))
            if kwargs.get('field_mapping') is not None:
                # todo: get tax based on geo
                res = wc_to_model.wp_parse_order(res, kwargs.get('field_mapping'), 
                                shipping_tax_id=kwargs.get('shipping_tax_id',2), 
                                item_tax_id=kwargs.get('item_tax_id', 1)) 
            return res

            

    @staticmethod        
    def get_products_at_shop(shop_type, shop_api, **kwargs):
        if shop_type == 'wc':
            data = wc.list_products(shop_api, 
                                params={
                                    'page': kwargs.get('page', 1), 
                                    'per_page': kwargs.get('batch_size', 10)
                                }) 
            for i, d in enumerate(data):
                data[i] = ShopConnector._parse_product(shop_type, shop_api, 
                                      d, kwargs.get('field_mapping'))
            return data
                
                            
    @staticmethod        
    def get_product_at_shop(shop_type, shop_api, product_id, **kwargs):
        if shop_type == 'wc':
            res = wc.get_product_by_id(shop_api, product_id) 
            return ShopConnector._parse_product(shop_type, shop_api, 
                                      res, kwargs.get('field_mapping'))
            
                

    @staticmethod    
    def get_category_taxonomy(shop_type, shop_api, **kwargs):
        if shop_type == 'wc':
            page = 1
            cats = []
            while True:
                batch_cats = wc.list_categories(shop_api, 
                                        page, 
                                        params=kwargs) 
                if batch_cats is None:
                    return None
                if len(batch_cats) == 0:
                    break
                cats += batch_cats 
                page += 1
            wc_to_model.wp_category_to_taxonomy(cats)
            return cats
    
    @staticmethod
    def update_customer_at_shop(shop_type, shop_api, destination_order_id, 
                             data, **kwargs):
        if shop_type == 'wc':
            model_to_wc.transform_customer(data)
            update = {}
            if kwargs.get('is_billing'):
                update['billing'] = data
            elif kwargs.get('is_shipping'):
                update['shipping'] = data
            else:
                logger.debug('need to specify biling or shipping')
                return 
            return wc.update_order(shop_api, destination_order_id, update) 

    @staticmethod
    def add_tracking(shop_type, shop_api, destination_order_id, 
                            message, **kwargs):
        if shop_type == 'wc':
            assert message.get('tracking_id')
            assert message.get('provider')
            tracking = f"tracking code {message['tracking_id']} service by \
                         {message['provider']}"
            res = wc.send_wc_message(shop_api,
                                destination_order_id,
                                tracking)
            return res

    @staticmethod
    def send_message(shop_type, shop_api, destination_order_id, 
                            message, **kwargs):
        if shop_type == 'wc':
            res = wc.send_wc_message(shop_api,
                                destination_order_id,
                                message)
            return res

    @staticmethod
    def _parse_product(shop_type, shop_api, product, field_mapping):
        if shop_type == 'wc':
            parsed = wc_to_model.parse_product(product, field_mapping)
            if parsed == 0:
                return 
            elif parsed == 1:
                variations = []
                for pid in product['variations']:
                    vari = wc.get_product_by_id(shop_api, pid) 
                    if vari: variations.append(vari)
                product['variations'] = variations 
                
                parsed = wc_to_model.parse_product_with_variations(product, field_mapping)
                
            return parsed



# -*- coding: utf-8 -*-

"""
WooCommerce API Class
"""

__title__ = "woocommerce-api"
__version__ = "3.0.0"
__author__ = "Claudio Sanches @ Automattic"
__license__ = "MIT"

from requests import request
from json import dumps as jsonencode
from time import time
from pypipet.plugins.woocommerce.oauth import OAuth
from requests.auth import HTTPBasicAuth
from urllib.parse import urlencode


class API(object):
    """ API Class """

    def __init__(self, url, consumer_key, consumer_secret, **kwargs):
        self.url = url
        self.consumer_key = consumer_key
        self.consumer_secret = consumer_secret
        self.wp_api = kwargs.get("wp_api", True)
        self.version = kwargs.get("version", "wc/v3")
        self.is_ssl = self.__is_ssl()
        self.timeout = kwargs.get("timeout", 800)
        self.verify_ssl = kwargs.get("verify_ssl", True)
        self.query_string_auth = kwargs.get("query_string_auth", False)
        self.user_agent = kwargs.get("user_agent", f"WooCommerce-Python-REST-API/{__version__}")

    def __is_ssl(self):
        """ Check if url use HTTPS """
        return self.url.startswith("https")

    def __get_url(self, endpoint):
        """ Get URL for requests """
        url = self.url
        api = "wc-api"

        if url.endswith("/") is False:
            url = f"{url}/"

        if self.wp_api:
            api = "wp-json"

        return f"{url}{api}/{self.version}/{endpoint}"

    def __get_oauth_url(self, url, method, **kwargs):
        """ Generate oAuth1.0a URL """
        oauth = OAuth(
            url=url,
            consumer_key=self.consumer_key,
            consumer_secret=self.consumer_secret,
            version=self.version,
            method=method,
            oauth_timestamp=kwargs.get("oauth_timestamp", int(time()))
        )

        return oauth.get_oauth_url()

    def __request(self, method, endpoint, data, params=None, **kwargs):
        """ Do requests """
        if params is None:
            params = {}
        url = self.__get_url(endpoint)
        auth = None
        headers = {
            "user-agent": f"{self.user_agent}",
            "accept": "application/json"
        }

        if self.is_ssl is True and self.query_string_auth is False:
            auth = HTTPBasicAuth(self.consumer_key, self.consumer_secret)
        elif self.is_ssl is True and self.query_string_auth is True:
            params.update({
                "consumer_key": self.consumer_key,
                "consumer_secret": self.consumer_secret
            })
        else:
            encoded_params = urlencode(params)
            url = f"{url}?{encoded_params}"
            url = self.__get_oauth_url(url, method, **kwargs)

        if data is not None:
            data = jsonencode(data, ensure_ascii=False).encode('utf-8')
            headers["content-type"] = "application/json;charset=utf-8"

        return request(
            method=method,
            url=url,
            verify=self.verify_ssl,
            auth=auth,
            params=params,
            data=data,
            timeout=self.timeout,
            headers=headers,
            **kwargs
        )

    def get(self, endpoint, **kwargs):
        """ Get requests """
        return self.__request("GET", endpoint, None, **kwargs)

    def post(self, endpoint, data, **kwargs):
        """ POST requests """
        return self.__request("POST", endpoint, data, **kwargs)

    def put(self, endpoint, data, **kwargs):
        """ PUT requests """
        return self.__request("PUT", endpoint, data, **kwargs)

    def delete(self, endpoint, **kwargs):
        """ DELETE requests """
        return self.__request("DELETE", endpoint, None, **kwargs)

    def options(self, endpoint, **kwargs):
        """ OPTIONS requests """
        return self.__request("OPTIONS", endpoint, None, **kwargs)



"""meta queries interfaces with Database"""
from .query import *
import logging
logger = logging.getLogger('__default__')
logger.setLevel(logging.DEBUG)


def get_server_timestamp(session):
    ts = query_server_timestamp(session)
    if ts:
        return ts[0]

def get_latest_record(table_obj, session, key:str, params: dict={}):
    return table_latest_record(table_obj, session, key,filters=params)

def search_exist(table_obj, session, params:dict):
    return  db_select(table_obj,
              session,
              filters=params)

def add_to_db(table_obj, session, obj):
    return db_insert(obj, table_obj, session)

def add_json_to_db(table_obj, session, data):
    res = db_insert_raw(table_obj, session, data)
    if res:
        return _object2dict(res)

def add_to_db_bulk(table_obj, session, objs:list):
    db_insert_bulk(objs, 
                table_obj,
                session)

def raw_query(query, session, params: dict={}):
    return db_raw_query(query, session, params)


def add_if_not_exist(table_obj, session, obj, unique_keys:list):
    res = db_select(table_obj, 
                   session, 
                   filters=obj.get_unique_attrs(unique_keys))
    if res and len(res) > 0:
        return res[0]

    return db_insert(obj, table_obj, session)

# def add_json_if_not_exist(table_obj, session, data):
#     res = db_select(table_obj, 
#                    session, 
#                    filters=data)
#     if len(res) > 0:
#         return res[0]
#     print(res)
#     return db_insert_raw(table_obj, session, data)

def update_data(table_obj, session, data:dict, params:dict):
    return db_update(table_obj, session, 
                   data, 
                   filters=params)

def update_bulk(table_obj, session, data:list):
    return db_update_bulk(table_obj, session, data)

def update_if_exit(table_obj, session, obj, unique_keys:list):
    filters = obj.get_unique_attrs(unique_keys)
    res = db_select(table_obj, 
                   session, 
                   filters=filters)
    if len(res) > 0:
        db_update(table_obj, session, 
                   obj.get_all_attrs(), 
                   filters=filters)
        return True 
    return False

def add_destination(table_obj, session, obj, unique_keys:list):
    filters = obj.get_unique_attrs(unique_keys)
    filters.update({
        'is_current_price': True
    })
    res = db_select(table_obj, 
                   session, 
                   filters=filters)
    if len(res) > 0:
        exist = res[0]
        return exist
        
    obj.set_attr('is_current_price', True)
    obj.set_attr('available', True)
    return db_insert(obj, table_obj, session)

def get_destination(table_obj, session, params:dict):
    params.update({
        'is_current_price': True
    })
    res = db_select(table_obj, 
                   session, 
                   filters=params)
    if len(res) == 0:
        return []
    if len(res) > 0:
        exist = [_object2dict(res[0])]
        return exist
    if len(res) > 1 and params.get('front_shop_id') is not None:
        logger.debug(f"sku {params.get('sku')} published multiple time \
        at shop {params['front_shop_id']}")


def update_destination_price(table_obj, session, price, params: dict):
    
    params.update({
         'is_current_price': True
    })
    res = db_select(table_obj, 
                   session, 
                   filters=params)
    if res is None:
        logger.debug(f'query failded {parmas}')
        return 

    if len(res) == 1:
        data = _object2dict(res[0])
        # print(data)
        #price no change
        if data['price'] == price:
            return res[0]

        db_update(table_obj, session, 
                   {'is_current_price': False}, 
                   filters=params)
        
        if data.get('id'):
            del data['id']
        data.update({
            'price': price,
            'is_current_price': True,
            'available': True
        })
       
        return db_insert_raw(table_obj, session, data)

    if len(res) > 1:
        logger.debug(f'published multiple times at \
                        same destination {parmas}')

def add_new_product(table_objs, session, product: object):   
    if product.category is not None:
        category = product.category
        unique_keys = list(category.get_all_attrs().keys())
        res = add_if_not_exist(table_objs.get(category.__table_name__), 
                    session, category, unique_keys)
        product.set_attr('category_id', res.id)

    # unique_keys = list(product.get_all_attrs().keys())
    unique_keys = ['identifier']
    res = add_if_not_exist(table_objs.get(product.__table_name__), 
                session, product, unique_keys)
    
    if res is None: return 
    product_id = res.id

    product.update_product_id(product_id)
    
    if len(product.variations) >0:
        variation_table = table_objs.get(product.variations[0].__table_name__)
        new_variations = []
        #check is_duplicate_skus
        for vari in product.variations:
            if _is_duplicate_sku(variation_table, session, vari.sku):
                logger.debug(f'duplicate sku {vari.sku}')
            else:
                new_variations.append(vari)
        db_insert_bulk(new_variations, 
                      variation_table,
                      session)
    return product

def get_product_with_variations(table_objs, session, params: dict, 
                                       include_published=False, front_shop_id=None, include_category=False):
    product_info = None
    if params.get('identifier') is not None:
        product_info = get_product_by_identifier(table_objs, session, 
                                          {'identifier': params['identifier']},
                                          include_category=include_category)
    elif params.get('product_id') is not None:
        product_info = get_product_by_identifier(table_objs, session, 
                                          {'id': params['product_id']},
                                          include_category=include_category) 
    elif params.get('sku') is not None:
        product_info = get_variation_info_by_sku(table_objs, session, params['sku'],
                                                 front_shop_id=front_shop_id)
        if product_info is None:
            logger.debug(f'invalid params {params}')
            return 
        if not include_published and front_shop_id and product_info.get('destinations'):
            publish_status, dest = is_sku_published(product_info.get('destinations'), 
                                params['sku'], front_shop_id)
            if publish_status == 1:
                return 
            elif publish_status == 2:
                product_info['set_available_flag'] = True
                product_info['destination_id'] = dest['id']
        return product_info

    if product_info is None:
        logger.debug(f'invalid params {params}')
        return 

    product_info = _tranfrom_product(table_objs.get('destination'), session,
                         product_info, front_shop_id, include_published=include_published)
    return product_info


def is_sku_published(dests, sku, front_shop_id):
    """
    return: (status_code, destination)
            status_code:
            0  not added yet
            1  added and available =True 
            2  added and available = False #need update the flag
    """
    for dest in dests:
        if dest['front_shop_id'] == front_shop_id and dest['sku'] == sku:
            if not dest['available']:
                logger.debug(f"sku {sku} has been add to shop {front_shop_id}, \
                need set availability flag as True") 
                return 2, dest
            logger.debug(f"sku {sku} has been add to shop {front_shop_id}") 
            return 1, dest
    return 0, None

def _tranfrom_product(table_obj, session, product_info, front_shop_id, include_published=False):
    """transform product_info depends on if it has multiple variations"""
    if product_info.get('variations') is None:
        return 
    if product_info.get('variations') and len(product_info['variations']) == 1:
        #return simple product dict 
        product_info.update(product_info['variations'])
        del product_info['variations'] 
        
        dests = get_destination(table_obj, session, {'sku': product_info['sku'],
                                                    'front_shop_id': front_shop_id})
        if dests is None:
            return
        if include_published:
            product_info['destinations'] = dests
            return product_info

        publish_status, dest = is_sku_published(dests, product_info['sku'], front_shop_id)
        if publish_status == 1:
            return
        elif publish_status == 2:
            product_info['set_available_flag'] = True
            product_info['destination_id'] = dest['id']

        return product_info
    
    parent_id = None
    for i, vari in enumerate(product_info['variations']):
        dests = get_destination(table_obj, session, {'sku': vari['sku'],
                                                    'front_shop_id': front_shop_id})
        
        if dests is None:
            logger.debug(f"destination query error sku {vari['sku']} at shop {front_shop_id}")
            return 
        if include_published:
            product_info['destinations'] = dests
        else:
            publish_status, dest = is_sku_published(dests, vari['sku'], front_shop_id)
            if publish_status == 1:
                product_info['variations'][0]['add_flag'] = True
                product_info['variations'][0]['destination_id'] = dest['id']
            elif publish_status == 2:
                product_info['variations'][0]['set_available_flag'] = True
                product_info['variations'][0]['destination_id'] = dest['id']

            if dest and dest.get('destination_parent_id'):
                parent_id = dest['destination_parent_id']
    if parent_id:
        product_info['parent_id'] = parent_id
    return product_info

# def get_destination_by_sku(table_objs, sku, front_shop_id):
#     params = {'sku': sku, 'is_current_price': True,
#               'front_shop_id': front_shop_id}
#     dests = db_select(table_objs.get('destination'), 
#                    session, 
#                    filters=params)
#     if len(dests) > 0:
#         return _object2dict(dests[0])
#     elif len(dests) > 1:
#         logger.debug(f'sku {sku} published multiple at shop {front_shop_id}')

def get_product_by_identifier(table_objs, session, params:dict, include_category=False):
    product = db_select(table_objs.get('product'), 
                        session, 
                        filters=params)
    if len(product) == 0:
        return None 
    
    product = product[0]
    variations =  db_select(table_objs.get('variation'), 
                   session, 
                   filters={'product_id': product.id})

    product_info = _object2dict(product)

    if include_category:
        category = db_select(table_objs.get('category'), 
                   session, 
                   filters={'id': product_info['category_id']})
        product_info['category'] = category[0].category

    if len(variations) > 0:
        product_info['variations'] = [_object2dict(v) for v in variations]
    
    if product_info.get('id'):
        del product_info['id']
    return product_info

def get_product_by_sku(table_objs, session, sku:str, include_category=False):
    variation = db_select(table_objs.get('variation'), 
                   session, 
                   filters={'sku': sku})
    if len(variation) == 0:
        return None 

    variation = _object2dict(variation[0])

    product = db_select(table_objs.get('product'), 
                   session, 
                   filters={'id': variation['product_id']})
    
    product_info = _object2dict(product[0])
    product_info.update(variation)

    if include_category:
        category = db_select(table_objs.get('category'), 
                   session, 
                   filters={'id': product_info['category_id']})
        product_info['category'] = category[0].category
        product_info['full_category'] = category[0].full_path

    return product_info

def get_product_name_by_sku(table_objs, session, sku:str):
    variation = db_select(table_objs.get('variation'), 
                   session, 
                   filters={'sku': sku})
    if len(variation) == 0:
        return None 

    variation = _object2dict(variation[0])

    product = db_select(table_objs.get('product'), 
                   session, 
                   filters={'id': variation['product_id']})
    
    product_info = _object2dict(product[0])
    product_info.update(variation)


    return product_info


def get_variation_info_by_sku(table_objs, session, sku:str, front_shop_id=None):
    variation = db_select(table_objs.get('variation'), 
                   session, 
                   filters={'sku': sku})
    if len(variation) == 0:
        return None 
   
    variations = db_select(table_objs.get('variation'), 
                   session, 
                   filters={'product_id': variation[0].product_id})
    
    product_info = {}

    if len(variations) > 1:
        #multi variation 
        assert front_shop_id is not None
        skus = [v.sku for v in variations]
        parent_id = get_destination_parent(table_objs.get('destination'), 
                                session, front_shop_id, skus)
        if parent_id:
            product_info['parent_id'] = parent_id
        product_info['variations'] = [_object2dict(variation[0])] # variation with the sku
        product_info['product_id'] = variations[0].product_id
    else:
        #single variation
        product_info.update(_object2dict(variation[0]))

    #variation = _object2dict(variation[0])
    
    # inv_table = table_objs.get('inventory')
    # inv =  get_inventory_by_sku(inv_table, session, sku)
    # if inv:
    #     variation['inventory'] = inv

    product = db_select(table_objs.get('product'), 
                   session, 
                   filters={'id': product_info['product_id']})
    
    product_info.update(_object2dict(product[0]))
    if product_info.get('id'):
        del product_info['id']
    #variation.update(product_info)

    category = db_select(table_objs.get('category'), 
                   session, 
                   filters={'id': product_info['category_id']})
    product_info['category'] = category[0].category
    # product_info['full_category'] = category[0].full_path
    
    #get existing destination
    params = {'sku': sku, 'is_current_price': True}
    if front_shop_id is not None:
        params['front_shop_id'] = front_shop_id
    dests = db_select(table_objs.get('destination'), 
                   session, 
                   filters=params)
    if len(dests) > 0:
        product_info['destinations'] = [_object2dict(d) for d in dests]

    return product_info

def get_product_by_destination(table_objs, session, params:dict):
    dest = db_select(table_objs.get('destination'), 
                   session, 
                   filters=params)
    if len(dest) == 0:
        return None 
    dest = dest[0]

    variation = db_select(table_objs.get('variation'), 
                   session, 
                   filters={'sku': dest.sku})
    if len(variation) == 0:
        return None 
    variation = variation[0]
    info = _object2dict(variation)

    product = db_select(table_objs.get('product'), 
                   session, 
                   filters={'id': variation.product_id})
    
    if len(product) == 0:
        return None 

    info.update(_object2dict(product[0]))
    return info

def get_order(table_objs, session, params:dict, **kwargs):
    order = db_select(table_objs.get('shop_order'), 
                   session, 
                   filters=params)
    if order is None or len(order) == 0:
        return None 
    order = order[0]
    order_info =  _object2dict(order)

    order_items = db_select(table_objs.get('order_item'), 
                   session, 
                   filters={'shop_order_id': order.id})
    if kwargs.get('item_info') == True:
        item_info =  []
        for d in order_items:
            prods = get_products(table_objs, session, {'sku': d.sku})
            
            item_info.append(prods[0]._asdict())
        
    else:
        order_items =  [_object2dict(d) for d in order_items]

    order_info['order_item'] = item_info

    billing = db_select(table_objs.get('customer'), 
                   session, 
                   filters={'id': order.billing_customer_id})
    if len(billing) > 0:
        order_info['billing_customer'] = _object2dict(billing[0]) 
        if order.billing_customer_id == order.shipping_customer_id:
            order_info['shipping_customer'] = order_info['billing_customer']
        else:
            shipping = db_select(table_objs.get('customer'), 
                   session, 
                   filters={'id': order.shipping_customer_id})
            if len(shipping) > 0:
                order_info['shipping_customer'] = _object2dict(shipping[0]) 
    
    return order_info



def aggregate_inventory_qty(table_objs, session, 
                       start_from_sku: str, params={}, 
                       batch_size=10,latest_hours=24):
    # res = get_inventory(table_objs, session, start_from_sku, batch_size,
    #                      params=params, latest_hours=latest_hours)
    res = get_latest_inventory_update(table_objs.get('inventory'), session, start_from_sku, batch_size,
                         params=params, latest_hours=latest_hours)
    if res is not None:
        return [r._asdict() for r in res]

def get_variation_instock_qty(table_objs, session, 
                       start_product_id: str, params={}, 
                       batch_size=10, latest_hours=24):
    res = get_variation_instock(table_objs, session, 
                         start_product_id, 
                         batch_size,
                         params=params,
                         latest_hours=latest_hours)
    
    if res is not None:
        return [r._asdict() for r in res]

# def get_inventory_update_after(table_objs, session, 
#                                     after_dt, params:dict={}):
#     res = get_inventory_update(table_objs, session, after_dt, params)
#     if res is not None:
#         return [r._asdict() for r in res]

def update_inventory(table_obj, session, 
              update_data: dict, params:dict):
    if update_data.get('currency') is None:
        update_data['currency'] = 'USD'
    res = db_select(table_obj, 
                   session, 
                   filters=params)
    if res and len(res) > 0:
        db_update(table_obj, session, 
                   update_data, 
                   filters={'id': res[0].id})
        return res[0]
    else:
        update_data.update(params)
        return db_insert_raw(table_obj, session, update_data)

# def get_inventory_by_sku(table_obj, session, sku:str):
#     res = db_sum_query(table_obj, 
#                  session, 
#                  table_obj.qty, 
#                  group_by=[table_obj.sku],
#                  params={'sku': sku})
#     if len(res) > 0:
#         return res[0]._asdict().get('sum')


def get_fulfilled_orders(table_objs, session, status='processing', 
                           shop_order_id_list=None, params:dict={}):
    if shop_order_id_list is None:
        ff = get_fulfilled_order_in_processing(table_objs, session, status=status, params=params)
        return [r._asdict() for r in ff]
    elif type(shop_order_id_list) is list:
        ff = get_fulfilled_order_in_processing(table_objs, session, status=status, params={
            'shop_order_id_list': shop_order_id_list
        })
        return ff
    else:
        logger.debug('invalid parmas')
        return []

def get_front_shop(table_obj, session, url:str=None, name:str=None):
    if name:
        res = db_select(table_obj, 
                   session, 
                   filters={'name': name})
        if len(res) > 0:
            return res[0] 

    if url:
        res = db_select(table_obj, 
                   session, 
                   filters={'url': url})
        if len(res) > 0:
            return res[0]  

def insert_list_raw(table_obj, session, data: list):
    """table_obj: sqlalchemy.ext.automap
       data: list of dict"""    
    for d in data:
        db_insert_raw(table_obj, session, d)

def _is_duplicate_sku(table_obj, session, sku):
    res = db_select(table_obj, session, filters={'sku': sku})
    if len(res) > 0:
        return True 

    return False

def _object2dict(table_obj):
    d = {}
    for k in table_obj.__table__.columns.keys():
        d[k] = table_obj.__dict__.get(k, None)
    return d



from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import inspect, text, Table,DateTime
from sqlalchemy.sql import func
import datetime, logging
from .base import *
import logging
logger = logging.getLogger('__default__')

def init_tables(engine, table_settings:dict):
    base = declarative_base()
    base.metadata.reflect(engine)

    for name, config in table_settings.items():
        if base.metadata.tables.get(name) is None:
            logging.debug(f"Creating table {name}")
            table = create_table(name, config, base.metadata)
            logging.debug(f'created table {name}')

    base.metadata.create_all(engine)

def drop_all_tables(table_settings, engine):
    #quickly drop tables in dev
    base = declarative_base()
    base.metadata.reflect(engine)
    tables = []
    for name, config in table_settings.items():
        table = base.metadata.tables.get(name)
        if table is not None:
            tables.append(table)

    for i in range(1, len(tables)+1):
        table = tables[-i]
        table.drop(engine)
    return tables



def create_table(table_name, config:dict, meta:declarative_base):
    """create a table
    (setting on db_setting.yaml)"""
    logger.debug(f'creating table {table_name}')
    args = []
    for col in config['columns']:
        args.append(get_column(col)) 

    # add created_at and updated_at
    args.append(
        Column('created_at',
         DateTime,  
        nullable=False, server_default=func.now())
    )
    args.append(
         Column('updated_at',
         DateTime, 
         onupdate=func.now(),
         default=func.now()
         )
    )
    #add constraints
    if config.get('unique_constraint'):
        for conf in config['unique_constraint']:
            args.append(get_unique_constraint(conf))
    if config.get('foreign_key_constraint'):
        args += get_foreignkey_constraint(config['foreign_key_constraint'])
  
    #add index
    if config.get('index'):
        for conf in config['index']:
            args.append(get_index(conf))
    
    return Table(table_name,
                  meta,
                  *args)




#!/usr/bin/python
#
# Copyright 2016 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Backing store implementation for OAuth2 clients based on config data."""

from __future__ import print_function
import json
import os
import sys

from pypipet.plugins.gg_merchant.shopping.content import _constants
import google.auth
import google.oauth2


class Storage(object):
  """Simple store for refresh token-based clients."""

  def __init__(self, config):
    self._config = config

  def _token_path(self):
    return os.path.join(self._config['path'], _constants.TOKEN_FILE)

  def get(self):
    """Attempt to retrieve the currently stored token.

    Returns:
      An instance of google.oauth2.credentials.Credentials if token
      retrieval succeeds, or None if it fails for any reason.
    """
    try:
      with open(self._token_path(), 'r') as infile:
        token = json.load(infile)
      client_info = retrieve_client_config(self._config)['installed']
      credentials = google.oauth2.credentials.Credentials(
          None,
          client_id=client_info['client_id'],
          client_secret=client_info['client_secret'],
          refresh_token=token['refresh_token'],
          token_uri=client_info['token_uri'],
          scopes=[_constants.CONTENT_API_SCOPE])
      # Access tokens aren't stored (and may be expired even if we did), so
      # we'll need to refresh to ensure we have valid credentials.
      try:
        credentials.refresh(google.auth.transport.requests.Request())
        print('Using stored credentials from %s.' % self._token_path())
        return credentials
      except google.auth.exceptions.RefreshError:
        print('The stored credentials in the file %s cannot be refreshed, '
              're-requesting access.' % self._token_path())
        return None
    except (IOError, ValueError, KeyError):
      return None

  def put(self, credentials):
    """Store the provided credentials into the appropriate file.

    Args:
      credentials: an instance of google.oauth2.credentials.Credentials.
    """
    token = {
        'refresh_token': credentials.refresh_token,
    }
    with open(self._token_path(), 'w') as outfile:
      json.dump(token, outfile, sort_keys=True, indent=2,
                separators=(',', ': '))


def retrieve_client_config(config):
  client_secrets_path = os.path.join(
      config['path'], _constants.CLIENT_SECRETS_FILE)
  with open(client_secrets_path, 'r') as json_file:
    client_config_json = json.load(json_file)
  if 'installed' not in client_config_json:
    print('Please read the note about OAuth2 client IDs in the '
          'top-level README.')
    sys.exit(1)
  return client_config_json



import click, logging
from .cli import cli
from pypipet.core.logging import setup_logging

from . import ( 
    init,
    catalog,
    product,
    template,
    order,
    inventory,
    fulfillment,
    static
)



import click 
from .cli import cli
from pypipet.core.operations.static import get_static_data
from pypipet.core.operations.static import add_static_data
from pypipet.core.operations.static import edit_static_data
from pypipet.core.fileIO.file_loader import read_csv_file

@cli.command()
@click.argument("action",  type=click.Choice(['add', 'edit', 'show']))
@click.option("--target", type=click.Choice(['supplier', 'front_shop', 'tax', 'category']))
@click.option("-f", "--filename",
                help=" add/edit  in bulk")
@click.pass_context
def Static(ctx, action, target, filename):
    """
    add/edit stastic data (supplier, tax, category, frontshop)
    """
    click.echo('{} {} ...'.format(action, target))

    project = ctx.obj['project']
    config = ctx.obj["config"]
    if config is None:
        click.secho('create project first',
                   fg='yellow')
        return

    project.initialize_project(config)
    table_classes = project.get_table_objects()
    sessionmaker = project.get_session_maker()
    session = sessionmaker()

    if action == 'add':
        _add_data(session, table_classes, target, filename)
    elif action == 'edit':
        _edit_data(session, table_classes, target, filename)
    elif action == 'show':
        _show_data(session, table_classes, target)
    else:
        click.secho('action not supported: {}'.format(action),
                fg='yellow')
    
    session.close()

def _show_data(session, table_classes, target):
    data = get_static_data(table_classes.get(target), session)
    if data is not None:
        for d in data:
            click.echo(d) 
    else:
        click.secho('can not get data: {}'.format(taret),
                fg='yellow')

def _add_data(session, table_classes, target, filename):
    data = _get_data_from_file(filename, target) 
    for d in data:
        click.echo('adding {} {}'.format(target, d))
        add_static_data(table_classes.get(target), session, d)


def _edit_data(session, table_classes, target, filename):
    data = _get_data_from_file(filename, target)
    for d in data:
        click.echo('editing {} {}'.format(target, d))
        edit_static_data(table_classes, session, d, target)


def _get_data_from_file(filename, target):
    data = read_csv_file(filename)
    res = []

    #validation
    for i, row in data.iterrows():
        if (target in ['supplier', 'front_shop', 'tax'] and row['name'].strip() == '') \
        or (target in ['category'] and row['category'].strip() == ''):
            click.secho('row {} missing name , \
             stop processing'.format(i+1), 
                        fg='yellow')
            return 
        res.append(dict(row))

    return res



from sqlalchemy import Column, Integer, String, Boolean
from sqlalchemy import DateTime, Float
from sqlalchemy import UniqueConstraint, ForeignKeyConstraint
from sqlalchemy import Index

data_type_mapping ={
    'string': String,
    'integer': Integer,
    'bool': Boolean,
    'datetime': DateTime,
    'float': Float
}

def get_column(col:dict) -> Column:
    kwargs = {} 
    if col.get('not_null') == True:
        kwargs['nullable'] = False
    if col.get('is_key') == True:
        kwargs['primary_key'] = True
    if col.get('default'):
        kwargs['default'] = col['default']
    try:
        if col['type'] == 'string' and col.get('string_len'):
            return Column(col['name'], 
                            data_type_mapping[col['type']](col['string_len']), 
                            **kwargs)
        return Column(col['name'], 
                    data_type_mapping[col['type']], 
                    **kwargs)
    except KeyError:
        print(f"key error at db setting | {col}")
        exit(0)
  
def get_unique_constraint(constraint:dict) -> UniqueConstraint:
    """
    :param constraints: dict of unique constraint setting,
                see db_setting.yaml.
    :return:   constraint object 
    """
    return UniqueConstraint(*constraint['cols'], 
                            name=constraint['name'])


def get_foreignkey_constraint(constraint:dict) -> [ForeignKeyConstraint]:
    """
    :param constraints: dict of foreign key constraint setting,
                see db_setting.yaml.
    :return: list of constraint object 
    """
    foreign_keys = []
    for key, val in constraint.items():
        foreign_keys.append(ForeignKeyConstraint([key],[val]))
    return foreign_keys

def get_index(index_config:dict) -> Index:
    """
    :param constraints: dict of index setting,
                see db_setting.yaml.
    :return:   index object 
    """
    name ='_'.join(['idx'] +index_config['cols'])
    if index_config.get('name') is not None:
        name = index_config['name']
    return Index(name, 
                 *index_config['cols'],
                 unique=index_config['is_unique'])





from pypipet.core.sql.query_interface import *
# from pypipet.core.shop_conn.gateway import *
from pypipet.core.model.product import Destination, Product, Variation
from .utility import get_front_shop_id
# from pypipet.core.transform.wc_to_model import parse_product
from copy import deepcopy

import logging
logger = logging.getLogger('__default__')

_DEBUG_ = True


def add_taxonomy_to_db(table_obj, session, shop_connector):
    categories = shop_connector.get_category_taxonomy(
                                            shop_connector.shop_type,
                                            shop_connector.shop_api
                                    )
    if categories:
        for cat in categories:
            db_insert_raw(table_obj, session, cat)
    else:
        logger.info('fetching taxonomy failed')


def load_products_from_shop(table_objs, session, shop_connector, 
                         start_from=1,  currency='USD'):
    start_page = start_from
    while True:
        logger.debug(f'page {start_page}')
        res = shop_connector.get_products_at_shop(
                                  shop_connector.shop_type, 
                                  shop_connector.shop_api, 
                                  batch_size=shop_connector.batch_size, 
                                  page=start_page,
                                  field_mapping=shop_connector.field_mapping)
        if res or len(res) == 0:
            logger.debug('done sync')
            break
        
        # if shop_connector.shop_type == 'wc':
        #     res =[parse_product(r, shop_connector.field_mapping)
        #               for r in res]
        add_product_to_db_bulk(table_objs, session, 
                                       shop_connector,res)
        start_page += 1
        if _DEBUG_ : break


def get_product_from_shop(table_objs, session, shop_connector, 
                                destination_product_id, currency='USD'):
    res = shop_connector.get_product_at_shop(
                                  shop_connector.shop_type, 
                                  shop_connector.shop_api, 
                                  destination_product_id,
                                  field_mapping=shop_connector.field_mapping
                                  )
    if res:
        return res

def add_product_to_db_bulk(table_objs, session, 
                            shop_connector,products:list, currency='USD'):
    for p in products:
        # logger.debug(p)
        if p is None: continue
        add_product_to_db(table_objs, session, p['product'])
        if p.get('destinations') is None \
                  or len(p['destinations']) == 0: 
            continue
        for dest in p['destinations']:
            #print(p['destinations'])
            add_destination_to_db(table_objs, 
                              session, 
                              shop_connector, 
                              dest,
                              currency=currency)

def add_product_to_db(table_objs, session, product_info: dict):
    if product_info is None: return 
    product = Product()
    #check category
    if product_info.get('category') is not None \
    and product_info.get('category').strip() != '' \
    and type(product_info.get('category')) is str:
        # print(product_info['product_name'])
        cat = _get_category_by_name(
                            table_objs.get('category'), 
                            session, 
                            product_info['category'])
        
        product_info.update(cat)
        del product_info['category']
    
    product.set_product(table_objs, product_info)
    
    return add_new_product(table_objs, session, product)


# def add_variation_to_db(table_objs, session, product_id: int,  variation_data: dict):
#     if variation_data.get('sku') is None:
#         logger.debug(f'missing {sku}')
#         return None

#     variation  = Variation() 
#     variation.set_variation(table_objs, variation_data)
    
#     #validate non-duplicate sku
#     exists = search_exist(table_objs.get(variation.__table_name__), 
#                           session, 
#                           {'sku': variation.sku})
#     if len(exists) > 0:
#         logger.debug(f'duplicate {variation.sku}')
#         return None

    
#     variation.set_attr('product_id', product_id)

#     res = add_to_db(table_objs.get(variation.__table_name__), 
#                      session, 
#                      variation)
#     if res is not None:
#         return variation


def add_destination_to_db(table_objs, session, shop_connector, 
                                        dest_data, currency='USD'):
    if shop_connector.front_shop_id is None:
        get_front_shop_id(table_objs, session, shop_connector)
    
    dest_data.update({
        'front_shop_id': shop_connector.front_shop_id,
        'is_current_price': True,
        'currency': currency
    })
    if dest_data.get('available') is None:
        dest_data['available'] = True 

    exist = is_exist_destination(table_objs.get('destination'), 
                                 session, dest_data)
    if exist:
        #update 
        #update_data(table_objs.get('destination'),  session,
                  #dest_data, {'id': dest_data['id']})
        pass
    else:
        #add new 
        dest = Destination()
        dest.set_destination(table_objs, dest_data)
        
        add_to_db(table_objs.get(dest.__table_name__), session, dest)

def is_exist_destination(table_obj, session, dest_data):
    dest = search_exist(table_obj, session,{
        'front_shop_id': dest_data['front_shop_id'],
        'sku': dest_data['sku']
    })
    if dest is None or len(dest) == 0:
        return False 
    else:
        for d in dest:
            if d.is_current_price:
                logger.debug(f'data exist id {d.id}')
                dest_data['id'] = d.id
                return True 
        return False



def get_product_info(table_objs, session, shop_connector, identifier=None,
                    sku=None, product_id=None, include_published=False,
                    include_category=False, **kwargs):
    params = None
    if identifier is not None:
        params = {'identifier': identifier}
    elif product_id is not None:
        params = {'product_id': product_id}
    elif sku is not None:
        params = {'sku': sku}
    if params is None:
        logger.debug('invalid product params')
        return 

    variation_info = get_product_with_variations(
                     table_objs, 
                     session, 
                     params, 
                     include_published=include_published,
                     include_category=include_category, 
                     front_shop_id=shop_connector.front_shop_id)

    return variation_info

def add_product_to_shop(table_objs, session, shop_connector, product_info, 
                                                    price=None, prices=None, **kwargs):
    if product_info.get('destinations'):
        logger.debug(f"has destinations, {product_info['destinations']}")
        return
    if (price is None and prices is None) or \
    (prices is not None and type(prices) is not dict):
        logger.debug('invalid product params')
        return 
    
    currency = kwargs.get('currency', 'USD')
    if product_info.get('variations') is None:
        #product without variations
        product_info['price'] = price
        product_info['currency'] = currency
        res = shop_connector.add_product_to_shop(shop_connector.shop_type, 
                                             shop_connector.shop_api, 
                                             product_info,
                                             attr_list=kwargs['attr_list']
                                             ) 
        if res is None:
            logger.debug(f"adding product to front shop failed {product_info['sku']}")
            return None
        #add to database
        res = add_json_to_db(table_objs.get('destination'), session, 
                    {
                        'sku': product_info['sku'],
                        'destination_product_id': res['id'],
                        'front_shop_id': shop_connector.front_shop_id,
                        'price': price,
                        'is_current_price': True,
                        'available': True,
                        'currency': currency
                    })
        return res
    else:
        for i, vari in enumerate(product_info['variations']):
            product_info['variations'][i]['currency'] = currency 
            if price: 
                product_info['variations'][i]['price'] = price 
            else:
                product_info['variations'][i]['price'] = \
                             prices.get(vari['sku'])
       
        assert kwargs.get('variation_attrs') is not None
        
        parent_id = None
        if kwargs.get('parent_id') is not None: parent_id = kwargs['parent_id']
        else: parent_id = product_info.get('parent_id')
        
        res = shop_connector.add_product_to_shop(shop_connector.shop_type, 
                                             shop_connector.shop_api, 
                                             product_info,
                                             parent_id=parent_id,
                                             **kwargs) # return dict to res
        
        if res is None or len(res) == 0:
            logger.debug(f"adding product to front shop failed {product_info['identifier']}")
            return None
   
        for i, vari in enumerate(product_info['variations']):
            if res.get(vari['sku']):
                if price is None:
                    price =  prices.get(vari['sku'])
                add_json_to_db(table_objs.get('destination'), session, 
                        {
                            'sku': vari['sku'],
                            'destination_product_id': res[vari['sku']]['id'],
                            'front_shop_id': shop_connector.front_shop_id,
                            'price': price,
                            'is_current_price': True,
                            'available': True,
                            'currency': currency,
                            'destination_parent_id': res[vari['sku']]['parent_id']
                        })
            else:
                logger.debug(f"missing results for sku {vari['sku']}")  
        return res  

def add_variation_to_shop(table_objs, session, shop_connector, 
                             sku:str, price:float, **kwargs):
    if shop_connector.front_shop_id is None:
        get_front_shop_id(table_objs, session, shop_connector)
    
    #transform data
    # variation_info = get_variation_info_by_sku(table_objs, session, sku, 
    #                  front_shop_id=shop_connector.front_shop_id)
    variation_info = get_product_with_variations(table_objs, session, 
                    {'sku': sku}, 
                     include_published=False, 
                     front_shop_id=shop_connector.front_shop_id)
    
    if variation_info is None:
        return 
    
    #check if it has been added but available flag need update
    if variation_info.get('set_available_flag'):
        #activate again 
        update_data(table_objs.get('destination'), 
                    session, 
                    {'available': True}, 
                    {'id': variation_info['destination_id'] })
        logger.info(f"variation sku {sku} reactivated")
            
        return 

    if variation_info.get('destinations') is not None:
        del variation_info['destinations']
    
    # if the sku is one of the variations of a product
    if variation_info.get('variations') and type(variation_info['variations']) is list:
        if variation_info.get('variations') is None or kwargs.get('ignore_variations'):
            variation_info.update(variation_info['variations'][0])
            del variation_info['variations']
        else:
            logger.info('add as variation')
            return add_product_to_shop(table_objs, session, shop_connector, 
                       variation_info, price=price, **kwargs)

    #add to frontshop 
    variation_info['price'] = price
    variation_info['currency'] = kwargs.get('currency', 'USD')
    res = shop_connector.add_product_to_shop(shop_connector.shop_type, 
                                             shop_connector.shop_api, 
                                             variation_info,
                                             **kwargs)
    # if res is None:
    #     #if exist, update
    #     res = _update_product_at_frontshop_by_sku(table_objs, session, 
    #                                       shop_connector, 
    #                                       sku, variation_info)
    #    
    if res is None:
        logger.debug(f'adding product to front shop failed {sku}')
        return None
    #add to database
    res = add_json_to_db(table_objs.get('destination'), session, 
                   {
                       'sku': sku,
                       'destination_product_id': res['id'],
                       'front_shop_id': shop_connector.front_shop_id,
                       'price': price,
                       'is_current_price': True,
                       'available': True,
                       'currency': variation_info['currency']
                   })
    return res


def get_cost_by_sku_from_db(table_obj, session, sku, params: dict={}):
    """apply when  skus have multiple suppliers"""
    params['sku'] = sku
    exists = search_exist(table_obj, session, params)
    if len(exists) > 0:
        prices = [inv.cost for inv in exists]
        return sum(prices)/len(prices)

def update_product_to_db(table_objs, session, product_info: dict):
    params = {}
    if product_info.get('id') is not None:
        params['id'] = product_info['id']
    elif product_info.get('identifier') is not None:
        params['identifier'] = product_info['identifier']
    else:
        logger.debug('missing product id and identifier')
        return 
    #check category
    if product_info.get('category') is not None \
    and product_info.get('category').strip() != '' \
    and type(product_info.get('category')) is str:
        # print(product_info['product_name'])
        cat = _get_category_by_name(
                            table_objs.get('category'), 
                            session, 
                            product_info['category'])
        
        product_info.update(cat)
        del product_info['category'] 
    
    if product_info.get('short_description'):
        product_info['short_description'] = product_info['short_description'][:1000]
    update_data(table_objs.get('product'), session, product_info, 
                params)    

def update_variation_to_db(table_obj, session, variation_info: dict):
    update_data(table_obj, session, variation_info, 
                {'sku': variation_info['sku']})     

def update_destination_to_db(table_obj, session, dest_info: dict, front_shop_id):
    params ={
        'sku': dest_info['sku'],
        'front_shop_id': front_shop_id,
        'is_current_price': True
    }
    if dest_info.get('price') is not None:
        destination = update_destination_price(table_obj, 
                                    session, dest_info['price'], params)
    
    update_data(table_obj, session, dest_info, params)  

def update_front_shop_price_bulk(table_objs, session, shop_connector, data: [dict]):
    if shop_connector.front_shop_id is None:
        get_front_shop_id(table_objs, session, shop_connector)
    for d in data:
        if d.get('price') is None: continue
        price = d['price']
        del d['price']

        d['front_shop_id'] = shop_connector.front_shop_id
        update_front_shop_price(table_objs, session, 
                            shop_connector, price, d)
    return data

def update_front_shop_price(table_objs, session, shop_connector, price:float, params:dict):
    if shop_connector.front_shop_id is None:
        get_front_shop_id(table_objs, session, shop_connector)
    #udpate database
    destination = update_destination_price(table_objs.get('destination'), 
                                              session, price, params)
    # print(destination.destination_product_id)
    if destination:
        #udpate front shop
        res = shop_connector.update_shop_product(
                            shop_connector.shop_type, 
                            shop_connector.shop_api,
                            {'price': price}, 
                            product_id=destination.destination_product_id,
                            parent_id=destination.destination_parent_id
                            )
        
        return res

def update_product_at_front_shop_bulk(table_objs, session, shop_connector, 
                                   data_list: list, batch_size=50,  **kwargs):
    update_batch = []
    

    for data in data_list:
        product_info = deepcopy({})
        product_id = None
        parent_id = None
        sku = data['variation']['sku']
        if data.get('destination') is None or \
        data['destination'].get('destination_product_id') is None:
            exists = search_exist(table_objs.get('destination'), 
                                session, 
                                {'sku': sku,
                                'front_shop_id': shop_connector.front_shop_id,
                                #    'available': True,
                                'is_current_price': True})
            if exists is None or len(exists) == 0:
                logger.info(f"variation sku {sku} not at shop or not available")
                continue
            product_id = exists[0].destination_product_id
            parent_id = exists[0].destination_parent_id
            if data['destination'].get('price') is None:
                data['destination']['price'] = exists[0].price
        else:
            product_id = data['destination']['destination_product_id']
            parent_id = data['destination'].get('destination_parent_id')
        
        product_info.update(data.get('product', {}))
        product_info.update(data.get('variation', {}))
        product_info.update(data.get('destination', {}))
        product_info['id'] = product_id
        product_info['parent_id'] = parent_id
        # print(product_info)
        update_batch.append(product_info)

        if len(update_batch) == batch_size:
            res_shop = shop_connector.update_shop_product_batch(
                    shop_connector.shop_type,
                    shop_connector.shop_api,
                    update_batch)
            # print(res_shop)
            #print(f'last update {product_id} ')
            update_batch = []

    if len(update_batch) > 0:
        res_shop = shop_connector.update_shop_product_batch(
                shop_connector.shop_type,
                shop_connector.shop_api,
                update_batch)
        # print(res_shop)
        print(f'last update {product_id} ')



def update_product_at_front_shop(table_objs, session, shop_connector, 
                                   data, sku, **kwargs):
    if shop_connector.front_shop_id is None:
        get_front_shop_id(table_objs, session, shop_connector)
    product_id = None
    parent_id = None
    product_info = {}
    if data.get('destination') is None or \
    data['destination'].get('destination_product_id') is None:
        exists = search_exist(table_objs.get('destination'), 
                              session, 
                              {'sku': sku,
                               'front_shop_id': shop_connector.front_shop_id,
                            #    'available': True,
                               'is_current_price': True})
        if exists is None or len(exists) == 0:
            logger.info(f"variation sku {sku} not at shop or not available")
            return None
        product_id = exists[0].destination_product_id
        parent_id = exists[0].destination_parent_id
        if data['destination'].get('price') is None:
            data['destination']['price'] = exists[0].price
    else:
        product_id = data['destination']['destination_product_id']
        parent_id = data['destination'].get('destination_parent_id')
    
    product_info.update(data.get('product', {}))
    product_info.update(data.get('variation', {}))
    product_info.update(data.get('destination', {}))
    #update frontshop 
    res = shop_connector.update_shop_product(
                                   shop_connector.shop_type, 
                                   shop_connector.shop_api, 
                                   product_info, 
                                   str(product_id), 
                                   parent_id=parent_id,
                                   **kwargs)
    logger.debug(f'{product_id}, {sku}')
    return res

def _get_category_by_name(table_obj, session, name):
    exists = search_exist(table_obj, session, {'category': name})
    if exists and len(exists) > 0:
        return {'category_id': exists[0].id}
    new_cat = add_json_to_db(table_obj, session, {'category': name})
    return {'category_id': new_cat['id']}

def _update_product_at_frontshop_by_sku(table_objs, session, shop_connector, 
                                          sku:str, data:dict):
    """search product at frontshop by sku, 
       if exist, update"""
    product_id = shop_connector.get_destination_product_id(
                                   shop_connector.shop_type, 
                                   shop_connector.shop_api,
                                   sku)
    if product_id is not None:
        return shop_connector.update_shop_product(shop_connector.shop_type, 
                                   shop_connector.shop_api, 
                                   data, 
                                   str(product_id),
                                   formated=True)


# from pypipet.core.shop_conn.gateway import *
from pypipet.core.model.order import Order
from pypipet.core.sql.query_interface import search_exist, get_order
from pypipet.core.sql.query_interface import add_if_not_exist, db_insert_bulk
from pypipet.core.sql.query_interface import get_destination, update_data
from pypipet.core.sql.query_interface import get_latest_record, add_json_to_db
from .utility import get_front_shop_id, _object2dict

import logging
logger = logging.getLogger('__default__')



def add_orders_to_db_bulk(table_objs, session, orders:list, shop_connector):
    if shop_connector.front_shop_id is None:
        get_front_shop_id(table_objs, session, shop_connector)
    
    for order in orders:
        if type(order) is dict:
            order_obj = Order()
            order_obj.set_order(table_objs, order)
        else:
            order_obj = order
        add_order_to_db(table_objs, session, order_obj,shop_connector.front_shop_id)


    session.close() 

def add_order_to_db(table_objs, session, order:Order, front_shop_id:int):
    #billing customer
    billing_customer = order.billing_customer
    if billing_customer:
        res = add_if_not_exist(table_objs.get(billing_customer.__table_name__), 
                        session, 
                        billing_customer,
                        unique_keys= ['email'])
        order.set_attr('billing_customer_id', res.id)

    #shipping customer
    shipping_customer = order.shipping_customer
    if shipping_customer:
        search_keys = ['first_name', 'last_name', 'postcode']
        if shipping_customer.get_attr('email') is not None:
            search_keys.append('email')
        res = add_if_not_exist(table_objs.get(shipping_customer.__table_name__), 
                            session, 
                            shipping_customer,
                            unique_keys=search_keys)
        order.set_attr('shipping_customer_id', res.id)
        # print(res.id)

    order.set_attr('front_shop_id', front_shop_id)

    res = search_exist(table_objs.get(order.__table_name__),
                       session, 
                       {'destination_order_id': order.destination_order_id})
    if len(res) > 0:
        order.update_order_id(res[0].id)
    else:
        res =add_json_to_db(table_objs.get(order.__table_name__), 
                        session, 
                        order.get_all_attrs())
        order.update_order_id(res['id'])
    
    if len(order.order_items) > 0:
        table_obj = table_objs.get(order.order_items[0].__table_name__)
        for item in order.order_items:
            dests = get_destination(
                table_objs.get('destination'),
                session,
                params = {
                    'sku': item.sku,
                    'front_shop_id': front_shop_id
                }
            )
            if dests is None or len(dests) == 0:
                logger.debug('no product found at destination shop')
                return None
            item.set_attr('destination_id', dests[0]['id'])

            #by default, ship all as request
            if item.get_attr('ship_qty') is None:
                item.set_attr('ship_qty', item.order_qty)

        db_insert_bulk(order.order_items, 
                      table_obj,
                      session)


    

def get_orders_from_shop(table_objs, session, shop_connector):
    if shop_connector.front_shop_id is None:
        get_front_shop_id(table_objs, session, shop_connector)
    
    previous_upddate = get_latest_record(
                        table_objs.get('shop_order'), 
                        session, 
                        key='destination_order_id',
                        params={'front_shop_id': shop_connector.front_shop_id})
    latest_order_id = None
    if previous_upddate:
        latest_order_id = previous_upddate.destination_order_id
    
    order_data = shop_connector.sync_shop_orders(
                    shop_connector.shop_type,
                    shop_connector.shop_api,
                    shop_connector.field_mapping,
                    latest_order_id = latest_order_id)
    
    # orders = []
    # for item in order_data:
    #         order = Order()
    #         order.set_order(table_objs, item)
    #         orders.append(order)
    return order_data

def get_order_from_shop_by_id(shop_order_id, shop_connector):
    order_info = shop_connector.get_order_at_shop(
                                   shop_connector.shop_type, 
                                   shop_connector.shop_api, 
                                   field_mapping=shop_connector.field_mapping,
                                   destination_order_id=shop_order_id)
    return order_info

def get_order_info(table_objs, session, shop_order_id, front_shop_id, item_info=True):
    order_info = get_order(table_objs, session, params={
         'front_shop_id': front_shop_id,
         'destination_order_id': shop_order_id
    }, item_info=item_info)
    return order_info

def get_order_by_status(table_obj, session, status, params:dict={}):
    params.update({'status': status})
    orders = search_exist(table_obj,
                          session,
                          params)
    if len(orders) > 0 :
        return [_object2dict(order) for order in orders]
    return orders




# def update_orders_to_db(table_objs, session, shop_connector, orders: list, **kwargs):
#     for i, order_info in enumerate(orders):
#         if order_info.get('destination_order_id') is None:
#             continue
        
#         order_update = get_order_from_shop_by_id(
#                          order_info['destination_order_id'], shop_connector)

#         order = Order()
#         order.set_order(table_objs, order_update)

#         order_info.update(order.get_all_attrs())
        
#         order_info['response'] = update_order_to_db(
#                            table_objs.get('shop_order'), 
#                            session, 
#                            order_info)

#         order_info.update(order_update)
#         orders[i] = order_info
#     return orders

def update_order_to_db(table_obj, session, order_info:dict):
    if order_info.get('id') is not None:
        shop_order_id = order_info['id']
        del order_info['id']
        return update_data(table_obj, 
                    session, 
                    order_info, 
                    {'id': shop_order_id})
        order_info['id'] = shop_order_id
    elif order_info.get('destination_order_id') is not None \
    and order_info.get('front_shop_id') is not None:
        return update_data(table_obj, 
                    session, 
                    order_info, 
                    {'destination_order_id': order_info['destination_order_id'],
                     'front_shop_id': order_info['front_shop_id']})
    else:
        logger.debug(f'invalid order identifier {order_info}')



# def update_order_items_from_shop(table_objs, session, shop_connector,
#                                  orders:[dict]):
#     for order_info in orders:
#         update_order_item_from_shop(table_objs, 
#                                      session, 
#                                      shop_connector, 
#                                      order_info)
#     return  orders

# def update_order_item_from_shop_to_db(table_objs, session, shop_connector, params: dict, **kwargs):
#     shop_order_id = params.get('shop_order_id')
#     destination_order_id = params.get('destination_order_id')

#     order_info = shop_connector.get_order_at_shop(
#                                    shop_connector.shop_type, 
#                                    shop_connector.shop_api, 
#                                    field_mapping=shop_connector.field_mapping,
#                                    destination_order_id=destination_order_id)
    
#     order = Order()
#     order.set_order(table_objs, order_info)
#     for item in order.order_items:
#         item.set_attr('shop_order_id', shop_order_id)
#         update_order_item_to_db(table_objs, session, shop_connector.front_shop_id, 
#                                    item.get_all_attrs(),  
#                                    item.get_unique_attrs(
#                                        ['shop_order_id', 
#                                         'destination_product_id']
#                                    ))

#     return order


def sync_order_change_from_shop(table_objs, session, shop_connector, destination_order_id):
    order_info = get_order_from_shop_by_id(destination_order_id, shop_connector)

    order = Order()
    order.set_order(table_objs, order_info)
    order.set_attr('front_shop_id', shop_connector.front_shop_id)
    update_order_to_db(table_objs.get('shop_order'), session, order.get_all_attrs())

    #get order.id
    order_id = None
    exist_order = search_exist(table_objs.get('shop_order'), session, 
                    {'destination_order_id': destination_order_id,
                    'front_shop_id': shop_connector.front_shop_id})
    if exist_order and len(exist_order) > 0:
        order_id = exist_order[0].id

    for item in order.order_items:
        update_order_item_to_db(table_objs, session, shop_connector.front_shop_id, 
                        item.get_all_attrs(),  
                        params={'destination_product_id': item.destination_product_id,
                                'shop_order_id': order_id})
    
def update_order_item_to_shop(shop_connector, destination_order_id, order_items:list):
    items = {}
    for item in order_items:
        if item.get('destination_product_id') is None:
            logger.debug(f'missing destination_product_id {item}')
            continue
        items[item['destination_product_id']] = item
    if len(items) == 0:
        return 
    shop_connector.update_order_item_at_shop(
                            shop_connector.shop_type, 
                            shop_connector.shop_api, 
                            destination_order_id,
                            items)


def update_order_item_to_db(table_objs, session, front_shop_id, 
                                   order_item: dict,  params: dict):
    exist_item = search_exist(table_objs.get('order_item'), session, params)
    if len(exist_item) > 0:
        exist_item = exist_item[0]
        update_data(table_objs.get('order_item'), session, 
                      order_item, {'id': exist_item.id})
    else:
        print('add new order item', order_item)
        dests = get_destination(
            table_objs.get('destination'),
            session,
            params = {
                'destination_product_id': order_item['destination_product_id'],
                'front_shop_id': front_shop_id
            }
        )
        if dests is None or len(dests) == 0:
            logger.debug('no product found at destination shop')
            return None
        order_item['destination_id'] = dests[0]['id']

        #by default, ship all as request
        if order_item.get('ship_qty') is None:
            order_item['ship_qty'] = order_item['order_qty']

        add_json_to_db(table_objs.get('order_item'), session, order_item)



def update_customer_to_db(table_obj, session, customer_info, **kwargs):
    params = {}
    if kwargs.get('id') is not None:
        params['id'] = kwargs['id']
    elif kwargs.get('email') is not None:
        params['email'] = kwargs['email']
    else:
        logger.debug('invalid customer id or email')
        return 

    update_data(table_obj, 
                session, 
                customer_info, 
                params)

def update_customer_to_shop( session, shop_connector, data, 
                    destination_order_id, is_billing=False,is_shipping=False):
    shop_connector.update_customer_at_shop(
                                    shop_connector.shop_type, 
                                    shop_connector.shop_api, 
                                    destination_order_id, 
                                    data, 
                                    is_billing=is_billing,
                                    is_shipping=is_shipping) 

    
  
def send_message_to_customer(shop_connector, destination_order_id, message):
    shop_connector.send_message(shop_connector.shop_type, 
                                shop_connector.shop_api, 
                                destination_order_id, 
                                message)




import click 
from .cli import cli
from pypipet.core.shop_conn.shop_connector import ShopConnector
from pypipet.core.operations.utility import get_front_shop_id
from pypipet.core.operations.order import get_orders_from_shop
from pypipet.core.operations.order import  get_order_from_shop_by_id
from pypipet.core.operations.order import  add_order_to_db
from pypipet.core.operations.order import  add_orders_to_db_bulk
from pypipet.core.operations.order import  update_order_to_db
from pypipet.core.operations.order import  get_order_info
from pypipet.core.sql.query_interface import search_exist
from pypipet.core.model.order import Order

from pypipet.core.fileIO.file_loader import read_csv_file
from .utility import col2dict_update
from pprint import pprint

_PROCESSING = 'processing'

@cli.command()
@click.argument("action", type=click.Choice(['sync', 'status', 'show',
                                           'add', 'edit']))
@click.option("--shop", 
                help="frontshop name (in setting.shops)")
@click.option("--id",
                help="order id at frontshop")
@click.option("--manual",type=click.Choice(['refunded', 'completed', 
         'cancelled']), help="manual input for status")
@click.option("-f", "--filename",
                help=" edit order attributes")
@click.pass_context
def order(ctx, action, shop, id, manual, filename):
    """
    add/edit orders (by skus or by file)
    """
    click.echo('{} orders at {}...'.format(action, shop))

    project = ctx.obj['project']
    config = ctx.obj["config"]
    if config is None:
        click.secho('create project first',
                   fg='yellow')
        return

    project.initialize_project(config)
    table_classes = project.get_table_objects()
    sessionmaker = project.get_session_maker()
    session = sessionmaker()
    shop_config = project.config.get('shops').get(shop)

    if shop_config is None:
        click.secho('shop setting not found: {}'.format(shop),
                fg='yellow')
        session.close()
        return 
    
    shop_conn = ShopConnector(shop, 
                                shop_config['site_type'],
                                batch_size=shop_config['batch_size']) 
    shop_conn.set_shop_api(project.get_wp_api(shop))
    project.set_shop_connector(shop, shop_conn)
    get_front_shop_id(table_classes, session, shop_conn)

    if action == 'sync':
        if id is not None:
            _sync(session, table_classes, shop_conn, order_id=id)
        else:
            _sync(session, table_classes, shop_conn)

    elif action == 'status':
        _status(session, table_classes, shop_conn, 
               manual, order_id=id)
    elif action == 'show':
        if id is None:
            click.echo('missing order_id')
            return 
        _get_order(session, table_classes,shop_conn, id)
    elif action == 'add':
        pass 
    elif action == 'edit':
        pass
    else:
        click.secho('action not supported: {}'.format(action),
                fg='yellow')
    
    session.close()


def _get_order(session, table_classes, shop_conn, order_id):
    order_info = get_order_info(table_classes, 
                                session, 
                                order_id, 
                                shop_conn.front_shop_id)
    click.echo(order_info)
    return order_info

def _sync(session, table_classes, shop_conn, order_id=None):
    if order_id is None:
        orders = get_orders_from_shop(table_classes, session, shop_conn)
        add_orders_to_db_bulk(table_classes, session, orders, shop_conn)
        click.echo('sync orders total {}'.format(len(orders)))
    else:
        order = Order()
        order.set_order(table_classes, 
                       get_order_from_shop_by_id(order_id, shop_conn))
        add_order_to_db(table_classes, session, order,
                                shop_conn.front_shop_id)
        click.echo('sync order {}'.format(order_id))    

def _status(session, table_classes, shop_conn, manual, order_id=None):
    if order_id is None:
        #update status from frontshop to db
        orders = search_exist(table_classes.get('shop_order'),
                              session,
                              {'status': _PROCESSING})
        update_n = 0
        for order in orders:
            click.echo('updating order {}'.format(order.destination_order_id))
            order_info = shop_conn.get_order_at_shop(
                            shop_conn.shop_type, 
                            shop_conn.shop_api, 
                            field_mapping=shop_conn.field_mapping,
                            destination_order_id=order.destination_order_id)
            if order_info['status'] != _PROCESSING:
                update_order_to_db(table_classes.get('shop_order'),
                            session,
                            {'destination_order_id': order_id,
                            'front_shop_id': shop_conn.front_shop_id,
                            'status': order_info['status']})
                update_n += 1

        click.echo('sync orders total {} out of  {}'.format(update_n, len(orders)))
    else:
        status = None
        if manual is None:
            #udpate order(with order #) status from frontshop to db
            order_info = shop_conn.get_order_at_shop(
                            shop_conn.shop_type, 
                            shop_conn.shop_api, 
                            field_mapping=shop_conn.field_mapping,
                            destination_order_id=order_id)
            status = order_info['status']
        else:
            shop_conn.update_order_status_at_shop(
                shop_conn.shop_type, 
                shop_conn.shop_api, 
                destination_order_id = order_id,
                status = manual
            )
            status = manual

        update_order_to_db(table_classes.get('shop_order'),
                            session,
                            {'destination_order_id': order_id,
                            'front_shop_id': shop_conn.front_shop_id,
                            'status': status})

        click.echo('update order {} status to {}'\
                  .format(order_id, status ))    



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pypipet',
 'pypipet.cli',
 'pypipet.core',
 'pypipet.core.fileIO',
 'pypipet.core.logging',
 'pypipet.core.model',
 'pypipet.core.operations',
 'pypipet.core.pipeline',
 'pypipet.core.shop_conn',
 'pypipet.core.sql',
 'pypipet.core.transform',
 'pypipet.plugins',
 'pypipet.plugins.canadapost',
 'pypipet.plugins.gg_merchant.shopping',
 'pypipet.plugins.gg_merchant.shopping.content',
 'pypipet.plugins.paypal',
 'pypipet.plugins.woocommerce']

package_data = \
{'': ['*'], 'pypipet.core': ['default_setting/*']}

install_requires = \
['click>=7.0,<8.0',
 'pandas>=1.2.0,<2.0.0',
 'pyyaml>=5.3.1,<6.0.0',
 'requests>=2.23.0,<3.0.0',
 'sqlalchemy>=1.4.27,<2.0.0']

entry_points = \
{'console_scripts': ['pypipet = pypipet.cli:cli']}

setup_kwargs = {
    'name': 'pypipet',
    'version': '0.0.14a0',
    'description': 'pypipet',
    'long_description': '[![GitHub license](https://img.shields.io/github/license/Naereen/StrapDown.js.svg)](https://github.com/Naereen/StrapDown.js/blob/master/LICENSE)\n[![Generic badge](https://img.shields.io/badge/Status-dev-orange.svg)](https://shields.io/)\n[![Maintenance](https://img.shields.io/badge/Maintained%3F-yes-green.svg)](https://GitHub.com/Naereen/StrapDown.js/graphs/commit-activity)\n[![Generic badge](https://img.shields.io/badge/Pypi-0.0.1a-blue.svg)](https://shields.io/)\n[![Generic badge](https://img.shields.io/badge/Python-3.8-blue.svg)](https://shields.io/)\n\n#### introduction\n\n**`PyPipet`** is an open source project, aiming to integrate data flows in online retailing. It simplifies the data pipeline of ecommerce, for example, adding catalog, update product, manage inventory and orders, etc. It is customized for small business who are selling on wordpress (for example, with woocommerce), shopify, ebay, amazon, Paypal, etc. It extremely handy if the business is selling on multiple platforms. PyPipet is under dev. The latest version 0.0.1a. \n\n* For source code,  visit  [github repositoty](https://github.com/pypipet/pypipet).\n* For documentation, vist [docs](https://pypipet.github.io/docs/)\n\n#### [dependencies](/dependencies)\n\n#### installation\n\n    pip install pypipet\n\n    or \n\n    pip3 install pypipet\n\n#### to-do list\n\n- [ ] connect to shopify\n- [ ] connect to bigcommerce\n- [ ] add email template\n- [ ] UI\n- [ ] connect to Google Analytics\n\n#### [quick start guide](https://pypipet.github.io/docs//quick_start/create_project/)',
    'author': 'pypipet and contributors',
    'author_email': 'pypipet@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from pypipet.core.sql.query_interface import search_exist


def get_front_shop_id(table_objs, session, shop_connector):
    front_shop = search_exist(
                            table_objs.get('front_shop'), 
                            session, 
                            {'name': shop_connector.shop_name})
    
    if front_shop is not None and len(front_shop):
        shop_connector.set_front_shop_id(front_shop[0].id)


def chunk_list(lst, n):
    for i in range(0, len(lst), n):
        yield lst[i:i+n]

def _object2dict(table_obj):
    d = {}
    for k in table_obj.__table__.columns.keys():
        d[k] = table_obj.__dict__.get(k, None)
    return d  


def flat_dict(d):
    flat_d = {}
    for key, val in d.items():
        if isinstance(val, dict):
            for sub_key, sub_val in val.items():
                flat_d[f'{key}.{sub_key}'] = sub_val 
        elif isinstance(val, list):
            for i, item in enumerate(val):
                for sub_key, sub_val in item.items():
                    flat_d[f'{key}.{sub_key}.{i}'] = sub_val   
        else:
            flat_d[key] = val 
    return flat_d
                
                  



from .action import ActionMap

import logging
logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

class Pipeline():
    def __init__(self, data: list, actions: list):
        self.actions = actions
        self.data = data

    def run(self):
        for action in self.actions:
            if action.name == 'external':
                self.data = action.run_external_func(self.data, action.params)
            else:
                print('pipeline action', action.name)
                data = ActionMap.mapping(self.data, action)
                self.data = data

   




from .base_object import BaseObject
from pprint import pprint
import logging
logger = logging.getLogger('__default__')


class Inventory(BaseObject):
    __table_name__ = 'inventory'
    def __init__(self, attrs: dict = {}):
        super().__init__(attrs)


    def set_inventory(self, table_objs, attrs: dict):
        self.set_attrs_by_table_class(table_objs.get(self.__table_name__), 
                                     attrs)

class Destination(BaseObject):
    __table_name__ = 'destination'
    def __init__(self, attrs: dict = {}):
        super().__init__(attrs)

    def set_destination(self, table_objs, attrs: dict):
        self.set_attrs_by_table_class(table_objs.get(self.__table_name__), 
                                     attrs)
    def print_destination(self):
        print('destinations')
        pprint(self.get_all_attrs())
    # def set_destination(self, attr_names: dict, attrs: dict):
    #     if attr_names.get('destination'):
    #         self.set_attrs(attr_names['destination'], attrs)

class Variation(BaseObject):
    __table_name__ = 'variation'
    def __init__(self, attrs: dict = {}):
        super().__init__(attrs)
        self.destinations = []

    # def get_variation(self):
    #     variation = self.get_all_attrs()
    #     if self.destinations:
    #         variation['destinations'] = []
    #         for dest in self.destinations:
    #             variation['destinations'].append(dest.get_destination())
    #     return variation


    def set_variation(self, table_objs, attrs: dict):
        self.set_attrs_by_table_class(table_objs.get(self.__table_name__), 
                                     attrs)
        if attrs.get('destinations'):
            for dest_data in attrs['destinations']:
                dest = Destination()
                dest.set_destination(
                                        table_objs, 
                                        dest_data)
                self.destinations.append(dest)

    def print_variation(self):
        print('variations')
        pprint(self.get_all_attrs())
        for dest in self.destinations:
            dest.print_destination()
    # def add_to_shop(self, front_shop:dict, data:dict):
    #     """"
    #     add variation to front shop
    #     front_shop: shop type and api 
    #     data: destination attrs
    #     """
    #     if self.sku is None:
    #         logger.debug('missing sku')
    #         return None 
    #     data['sku'] = self.sku

    #     if data.get('price') is None:
    #         logger.debug('missing price')
    #         return None
    #     if data.get('destination_product_id') is None \
    #                                 and front_shop is None:
    #         logger.debug('missing  product id at front_shop')
    #         return None

    #     if data.get('destination_product_id') is None:
            # data['destination_product_id'] = get_destination_product_id(
    #             front_shop, self.sku
    #         )

    #     destination = Destination(data)



class Product(BaseObject):
    __table_name__ = 'product'
    def __init__(self, attrs: dict = {}):
        super().__init__(attrs)
        self.variations = []
        self.category = None

    def get_product(self):
        product = self.get_all_attrs() 
        if self.variations:
            product['variations'] = []
            for vari in self.variations:
                product['variations'].append(vari.get_variation())
        return product

    def set_product(self, table_objs, attrs: dict):
        """
        set up product object
        :table_objs: table classes
        :attrs: dict - data dict
        """
        self.set_attrs_by_table_class(table_objs.get(self.__table_name__), 
                                     attrs)
        if attrs.get('category'):
            category = Category()
            category.set_category(table_objs, 
                                attrs['category'])
            self.category = category
        if attrs.get('variations'):
            for vari in attrs['variations']:
                variation = Variation()
                if vari.get('sku'):
                    vari['sku'] = str(vari['sku'])
                variation.set_attrs_by_table_class(
                                        table_objs.get(variation.__table_name__), 
                                        vari)
                self.variations.append(variation)

    def update_product_id(self, product_id):
        self.set_attr('id', product_id)
        for vari in self.variations:
            vari.set_attr('product_id', product_id)

    def print_product(self):
        print('product')
        pprint(self.get_all_attrs())
        for vari in self.variations:
            vari.print_variation()

     
class Category(BaseObject):
    __table_name__ = 'category'
    def __init__(self, attrs: dict = {}):
        super().__init__(attrs)

    def get_parent(self, sep='>') -> str:
        """
        :param sep: seperator in full path.
        :return:   parent category: str 
        """
        if self.full_path:
            categories = self.full_path.split(sep)
            if len(categories) < 2:
                return None
            self.set_attr('parent', categories[-2].strip())
            return self.parent

        return None   

    def validate(self, sep='>'):
        """
        :param sep: seperator in full path.
        :return:   parent category: str 
        """
        if self.category and self.category.strip() != '':
            return True 
        if self.full_path and self.full_path.strip() != '':
            self.get_parent(sep=sep)
        return False   

    def set_category(self, table_objs, attrs: dict):
        if type(attrs) is dict:
            self.set_attrs_by_table_class(table_objs.get(self.__table_name__), 
              attrs)
        elif type(attrs) is str:
            self.category = attrs


#to do remove this class
class Tax(BaseObject):
    def __init__(self, attrs: dict = {}):
        super().__init__(attrs)
        
    def override_tax_rate(self, rate):
        self.set_attr('override_rate', rate)
            
class FrontShop(BaseObject):
    def __init__(self, attrs: dict = {}):
        super().__init__(attrs)
        api_conn = None

    def get_api(self):
        pass

#to do remove this class
class SupplierInvoice(BaseObject):
    def __init__(self, attrs: dict = {}):
        super().__init__(attrs)

    def get_supplier(self):
        pass

    def set_invoice(self, attr_names: dict, attrs: dict):
        if attr_names.get('supplier_invoice'):
            self.set_attrs(attr_names['supplier_invoice'], attrs)


from pypipet.plugins.paypal import API
from requests import request
from requests.auth import HTTPBasicAuth


def get_api(url, client_id, secret):
    return API(url, client_id, secret)


def update_tracking(pp_api, tracking_data:list):
    endpoint = f"shipping/trackers-batch"
    res = pp_api.post(endpoint, data={'trackers':tracking_data})
    return res


#!/usr/bin/python
#
# Copyright 2016 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Various constants used in the Content API for Shopping samples."""

# Constants for configuration
CONFIG_FILE = 'merchant-info.json'
TOKEN_FILE = 'stored-token.json'
APPLICATION_NAME = 'Content API for Shopping Samples'

# Constants for authentication
CLIENT_SECRETS_FILE = 'client-secrets.json'
SERVICE_ACCOUNT_FILE = 'service-account.json'

# Constants needed for the Content API
SERVICE_NAME = 'content'
SERVICE_VERSION = 'v2.1'
SANDBOX_SERVICE_VERSION = 'v2.1sandbox'
CONTENT_API_SCOPE = 'https://www.googleapis.com/auth/' + SERVICE_NAME

# These constants define the identifiers for all of our example products/feeds.
#
# The products will be sold online.
CHANNEL = 'online'
# The product details are provided in English.
CONTENT_LANGUAGE = 'en'
# The products are sold in the USA.
TARGET_COUNTRY = 'US'

# Environment variable used for testing against different endpoints.
ENDPOINT_ENV_VAR = 'GOOGLE_SHOPPING_SAMPLES_ENDPOINT'
